"""Provides a reporter for generating graphical depictions of
resampling parent trees.
"""

from collections import namedtuple

import numpy as np
import networkx as nx
from matplotlib import cm

from geomm.free_energy import free_energy

from wepy.reporter.reporter import ProgressiveFileReporter
from wepy.analysis.network_layouts.tree import ResamplingTreeLayout
from wepy.analysis.network_layouts.layout_graph import LayoutGraph

from wepy.analysis.parents import resampling_panel, \
                                  parent_panel, net_parent_table,\
                                  ParentForest


class ResTreeReporter(ProgressiveFileReporter):
    """Reporter that generates resampling parent trees in the GEXF
    format."""

    FILE_ORDER = ('gexf_restree_path',)

    SUGGESTED_EXTENSIONS = ('restree.gexf',)

    MAX_PROGRESS_NORM = 1.0

    DISCONTINUOUS_NODE_SHAPE = 'square'
    """The shape of nodes that signify a discontinuity in the lineage."""

    DEFAULT_NODE_SHAPE = 'disc'
    """The shape of normal nodes that signify a continuity in the lineage."""

    def __init__(self,
                 resampler=None,
                 boundary_conditions=None,
                 row_spacing=None,
                 step_spacing=None,
                 default_node_radius=None,
                 progress_key=None,
                 max_progress_value=None,
                 colormap_name='plasma',
                 **kwargs):
        """Constructor for the ResTreeReporter.

        Parameters
        ----------

        resampler : Resampler
            Used to generate parental relations from resampling
            records.

        boundary_condition : BoundaryCondition, optional
            Used to determine discontinuities in the lineages.
             (Default value = None)

        row_spacing : float, default: 5.0
            Spacing between nodes in a single row in layout.

        step_spacing : float, default: 20.0
            Spacing between the rows of nodes in each step of the layout.

        default_node_radius : float, default: 1.0
            Default node radius to use.

        progress_key : str
            The key of the value in the progress records to use for
            coloring nodes in the tree.

        max_progress_value : float or None
            The maximum value to consider for progress values, if None
            no max value will be used.

        colormap_name : str, default: 'plasma'
            The name of the colormap to use from the matplotlib
            colormap library (i.e. 'matplotlib.cm.get_cmap')

        """

        assert resampler is not None, \
            "Must provide a resampler, this is used to get the correct records "\
            "from resampling data and is not saved in this object"

        assert boundary_conditions is not None, "must give the boundary condition class"

        # TODO: this should be made optional
        assert progress_key is not None, "Currently, a progress key must be given."

        super().__init__(**kwargs)

        # parameters for the layout
        self.row_spacing = row_spacing
        self.step_spacing = step_spacing
        self.default_node_radius = default_node_radius

        # this is the key we will use to get values for the progress out
        self.progress_key = progress_key
        self.max_progress_value = max_progress_value
        # get the actual colormap that will be used from matplotlib
        self.colormap = cm.get_cmap(name=colormap_name)

        # we need the decision class for getting parent child
        # relationships
        self._decision_class = resampler.decision

        # get the fields for the records of the resampling records, we
        # don't need to store the whole resampler
        self._resampling_record_field_names = resampler.resampling_record_field_names()

        # we also will need to know the shape and dtype of them
        self._resampling_field_names = resampler.resampling_field_names()
        self._resampling_field_shapes = resampler.resampling_field_shapes()
        self._resampling_field_dtypes = resampler.resampling_field_dtypes()


        # make a namedtuple record for these records
        # self._ResamplingRecord = namedtuple('{}_Record'.format('Resampling'),
        #                 ['cycle_idx'] + list(self._resampling_record_field_names))

        # we do the same for the bounadry condition class, although we
        # also need the class itself
        self._bc = boundary_conditions
        self._warping_record_field_names = self._bc.warping_record_field_names()

        self._warping_field_names = self._bc.warping_field_names()
        self._warping_field_shapes = self._bc.warping_field_shapes()
        self._warping_field_dtypes = self._bc.warping_field_dtypes()

        # self._WarpingRecord = namedtuple('{}_Record'.format('Warping'),
        #                             ['cycle_idx'] + list(self._warping_record_field_names))



        # initialize the parent table that will be generated as the
        # simulation progresses
        self._parent_table = []

        # the node ids of the warped nodes (ones with -1 parents)
        self._discontinuous_nodes = []

        # also keep track of the weights of the walkers
        self._walker_weights = []

        # and the progress values for each walker
        self._walkers_progress = []

    @property
    def parent_table(self):
        """The net parent table datastructure."""
        return self._parent_table

    def _make_resampling_record(self, record_d, cycle_idx):
        """Make a namedtuple resampling record from a dictionary
        representation.

        Parameters
        ----------
        record_d : dict of str : value
            The dictionary of values to make a namedtuple record from.
        cycle_idx : int
            The cycle index the record is associated with.

        Returns
        -------

        record : namedtuple

        """

        # make a namedtuple record for these records
        ResamplingRecord = namedtuple('{}_Record'.format('Resampling'),
                        ['cycle_idx'] + list(self._resampling_record_field_names))


        record = self._make_record(record_d, cycle_idx,
                                   self._resampling_record_field_names,
                                   self._resampling_field_names, self._resampling_field_shapes,
                                   ResamplingRecord)

        return record

    def _make_warping_record(self, record_d, cycle_idx):
        """Make a namedtuple warping record from a dictionary representation.

        Parameters
        ----------
        record_d : dict of str : value
            The dictionary of values to make a namedtuple record from.
        cycle_idx : int
            The cycle index the record is associated with.

        Returns
        -------

        record : namedtuple

        """

        # make a namedtuple record for these records
        WarpingRecord = namedtuple('{}_Record'.format('Warping'),
                                   ['cycle_idx'] + list(self._warping_record_field_names))

        record = self._make_record(record_d, cycle_idx,
                                   self._warping_record_field_names,
                                   self._warping_field_names, self._warping_field_shapes,
                                   WarpingRecord)

        return record


    @staticmethod
    def _make_record(record_d, cycle_idx,
                     record_field_names, field_names, field_shapes, rec_namedtuple):
        """Generic record making function.

        Parameters
        ----------
        record_d : dict of str : value

        cycle_idx : int

        record_field_names : list of str
            The record field names to use for the record.

        field_names : list of str
            All of the field names available for the field type.

        field_shapes : list of tuple of int
            Shapes for each field

        rec_namedtuple : namedtuple object
            The namedtuple class to make records out of

        Returns
        -------

        record : namedtuple object

        """

        # go through each field of the record
        rec_d = {'cycle_idx' : cycle_idx}
        for field_name, field_value in record_d.items():

            field_idx = field_names.index(field_name)
            field_shape = field_shapes[field_idx]

            # if it is not part of the record we just keep going and
            # don't bother
            if field_name not in record_field_names:
                continue

            # otherwise we need to do some formatting
            else:

                # if it is variable length (shape is Ellipsis) or if
                # it has more than one element cast all elements to
                # tuples
                if field_shape is Ellipsis:
                    value = tuple(field_value)

                # if it is not variable length make sure it is not more than a
                # 1D feature vector
                elif len(field_shape) > 2:
                    raise TypeError(
                        "cannot convert fields with feature vectors more than 1 dimension,"
                        " was given {} for {}".format(
                            field_value.shape[1:], field_name))

                # if it is only a rank 1 feature vector and it has more than
                # one element make a tuple out of it
                elif field_shape[0] > 1:
                    value = tuple(field_value)

                # otherwise just get the single value instead of keeping it as
                # a single valued feature vector
                else:
                    value = field_value[0]

            rec_d[field_name] = value

        # after you make the dictionary of this convert it to a
        # namedtuple record
        record = rec_namedtuple(*(rec_d[key] for key in rec_namedtuple._fields))

        return record


    def report(self, cycle_idx=None,
               resampled_walkers=None,
               warp_data=None,
               progress_data=None,
               resampling_data=None,
               **kwargs):
        """Generate the resampling tree GEXF file.

        Parameters
        ----------
        cycle_idx : int

        resampled_walkers : list of Walker

        warp_data : dict of str : value

        progress_data : dict of str : value

        resampling_data : dict of str : value

        """


        # we basically want to generate a new parent table from the
        # records we were just given and then add that to our existing
        # parent table

        # we first have to generate the resampling and warping records
        # though adding the cycle_idx

        resampling_records = [self._make_resampling_record(rec_d, cycle_idx)
                              for rec_d in resampling_data]

        warping_records = [self._make_warping_record(rec_d, cycle_idx)
                              for rec_d in warp_data]

        # tabulate the discontinuities
        for warping_record in warping_records:
            # if this record classifies as discontinuous
            if self._bc.warping_discontinuity(warping_record):
                # then we save it as one of the nodes that is discontinuous
                disc_node_id = (warping_record.cycle_idx, warping_record.walker_idx)
                self._discontinuous_nodes.append(disc_node_id)

        # get the weights of the resampled walkers since we will want
        # to plot them
        walker_weights = [walker.weight for walker in resampled_walkers]
        # add these to the collected weights
        self._walker_weights.append(walker_weights)

        # get the progress values for all of the walkers
        walkers_progress = [progress for progress in progress_data[self.progress_key]]
        self._walkers_progress.append(walkers_progress)

        # so we make a resampling panel from the records

        # TODO: this uses the cycle_idx in the records and thus wehn
        # you receive just one cycle it adds a empty records to the
        # beginning of it and only the last entry is the one we are
        # looking for so for now we just get it, and wrap back in a list
        res_panel = [resampling_panel(resampling_records, is_sorted=False)[-1]]

        # then the parent panel, and then the net parent table
        par_panel = parent_panel(self._decision_class, res_panel)
        cycle_parent_table = net_parent_table(par_panel)

        # add these to the main parent table
        self._parent_table.extend(cycle_parent_table)

        # now use this to create a new ParentForest without a WepyHDF5
        # file view. We exclude the discontinuities on purpose because
        # we want to keep the paren child relationships through the
        # warps for visualization and will just change the shapes of
        # the nodes to communicate the warping
        parent_forest = ParentForest(parent_table=self._parent_table)

        # The only ones that should be none are the roots which will
        # be 1.0 / N
        n_roots = len(self._parent_table[0])
        root_weight = 1.0 / n_roots

        # compute the size of the nodes to plot, which is the
        # free_energy. To compute this we flatten the list of list of
        # weights to a 1-d array then unravel it back out

        # put the root weights at the beginning of this array, and
        # account for it in the flattening
        flattened_weights = [root_weight for _ in self._walker_weights[0]]
        cycles_n_walkers = [len(self._walker_weights[0])]
        for cycle_weights in self._walker_weights:
            cycles_n_walkers.append(len(cycle_weights))
            flattened_weights.extend(cycle_weights)

        # now compute the free energy
        flattened_free_energies = free_energy(np.array(flattened_weights))

        # and ravel them back into a list of lists
        free_energies = []
        last_index = 0
        for cycle_n_walkers in cycles_n_walkers:
            free_energies.append(flattened_free_energies[last_index:last_index + cycle_n_walkers])
            last_index += cycle_n_walkers

        # we want to set the colors based on the progress, first we
        # scale all the progress values to a 1-100 scale
        if self.max_progress_value is not None:
            norm_ratio = self.MAX_PROGRESS_NORM / self.max_progress_value

        # otherwise we use the maximum value from the full progress
        # values dataset so far
        else:
            max_progress_value = max([max(row) for row in self._walkers_progress])
            # then use that for the ratio
            norm_ratio = self.MAX_PROGRESS_NORM / max_progress_value

        # now that we have ratios for normalizing values we apply that
        # and then use the lookup table for the color bar to get the
        # RGB color values
        colors = []
        for progress_row in self._walkers_progress:
            color_row = [self.colormap(progress * norm_ratio, bytes=True)
                         for progress in progress_row]
            colors.append(color_row)

        # put these into the parent forest graph as node attributes,
        # so we can get them out as a lookup table by node id for
        # assigning to nodes in the layout. We skip the first row of
        # free energies for the roots in this step
        parent_forest.set_attrs_by_array('free_energy', free_energies[1:])
        parent_forest.set_attrs_by_array('color', colors)

        # get the free energies out as a dictionary, flattening it to
        # a single value. Set the nodes with None for their FE to a
        # small number.

        # now we make the graph which will contain the layout
        # information
        layout_forest = LayoutGraph(parent_forest.graph)


        # get the free energy attributes out using the root weight for
        # the nodes with no fe defined

        # start it with the root weights
        node_fes = {(-1, i) : free_energy for i, free_energy in enumerate(free_energies[0])}
        for node_id, fe_arr in layout_forest.get_node_attributes('free_energy').items():

            if fe_arr is None:
                node_fes[node_id] = 0.0
            else:
                node_fes[node_id] = fe_arr

        # the default progress for the root ones is 0 and so the color
        # is also 0
        node_colors = {}
        for node_id, color_arr in layout_forest.get_node_attributes('color').items():

            if color_arr is None:
                # make a black color for these
                node_colors[node_id] = tuple(int(255) for a in range(4))
            else:
                node_colors[node_id] = color_arr

        # make a dictionary for the shapes of the nodes
        node_shapes = {}
        # set all the nodes to a default 'disc'
        for node in layout_forest.viz_graph.nodes:
            node_shapes[node] = self.DEFAULT_NODE_SHAPE
        # then set all the discontinuous ones
        for discontinuous_node in self._discontinuous_nodes:
            node_shapes[discontinuous_node] = self.DISCONTINUOUS_NODE_SHAPE

        # also set the color to black
        for discontinuous_node in self._discontinuous_nodes:
            node_colors[discontinuous_node] = tuple(int(0) for a in range(4))


        # we are going to output to the gexf format so we use the
        # pertinent methods

        # we set the sizes
        layout_forest.set_node_gexf_sizes(node_fes)

        # and set the colors based on the progresses

        layout_forest.set_node_gexf_colors_rgba(node_colors)

        # then set the shape of the nodes based on whether they were
        # warped or not
        layout_forest.set_node_gexf_shape(node_shapes)


        # now we get to the part where we make the layout (positioning
        # of nodes), so we parametrize a layout engine
        tree_layout = ResamplingTreeLayout(row_spacing=self.row_spacing,
                                           step_spacing=self.step_spacing,
                                           node_radius=self.default_node_radius)


        node_coords = tree_layout.layout(parent_forest,
                                         node_radii=node_fes)

        layout_forest.set_node_gexf_positions(node_coords)


        # now we can write this graph to the gexf format to the file
        nx.write_gexf(layout_forest.viz_graph, self.gexf_restree_path)



import numpy as np

from wepy.resampling.resamplers.resampler import Resampler, ResamplerError
from wepy.resampling.decisions.clone_merge import MultiCloneMergeDecision


class CloneMergeResampler(Resampler):
    """Abstract base class for resamplers using the clone-merge decision
    class.

    Provides some common functionality including handling minimum and
    maximum numbers of walkers and checking constraints on that.

    The 'assign_clones' method is a convenience method for generating
    decision records from two intermediate representation data
    structures which are convenient to produce: merge_groups and
    walker_clone_nums. See the docstring for more details on the
    format of these.

    """

    DECISION = MultiCloneMergeDecision

    RESAMPLING_FIELDS = DECISION.FIELDS + Resampler.CYCLE_FIELDS
    RESAMPLING_SHAPES = DECISION.SHAPES + Resampler.CYCLE_SHAPES
    RESAMPLING_DTYPES = DECISION.DTYPES + Resampler.CYCLE_DTYPES

    RESAMPLING_RECORD_FIELDS = DECISION.RECORD_FIELDS + Resampler.CYCLE_RECORD_FIELDS

    def __init__(self, pmin=1e-12, pmax=0.1,
                 min_num_walkers=Ellipsis,
                 max_num_walkers=Ellipsis,
                 **kwargs):
        """Constructor for CloneMegerResampler class.

        Parameters
        ----------

        pmin : float
            The minimum probability any walker is allowed to have.

        pmax : float
            The maximum probability any walker is allowed to have.

        """

        super().__init__(min_num_walkers=min_num_walkers,
                         max_num_walkers=max_num_walkers,
                         **kwargs)

        self._pmin=pmin
        self._pmax=pmax

    @property
    def pmin(self):
        return self._pmin

    @property
    def pmax(self):
        return self._pmax


    def _init_walker_actions(self, n_walkers):
        """Returns a list of default resampling records for a single
        resampling step.

        Parameters
        ----------

        n_walkers : int
            The number of walkers to generate records for

        Returns
        -------

        decision_records : list of dict of str: value
            A list of default decision records for one step of
            resampling.

        """

        # determine resampling actions
        walker_actions = [self.decision.record(
                                enum_value=self.decision.default_decision().value,
                                target_idxs=(i,))
                    for i in range(n_walkers)]

        return walker_actions

    def _check_resampled_walkers(self, resampled_walkers):
        """Check constraints on resampled walkers.

        Raises errors when constraints are violated.

        Parameters
        ----------
        resampled_walkers : list of Walker objects

        """

        walker_weights = np.array([walker.weight for walker in resampled_walkers])

        # check that all of the weights are less than or equal to the pmax
        overweight_walker_idxs = np.where(walker_weights > self.pmax)[0]
        if len(overweight_walker_idxs) > 0:

            raise ResamplerError("All walker weights must be less than the pmax, "
                                 "walkers {} are all overweight".format(
                                     ','.join([str(i) for i in overweight_walker_idxs])))

        # check that all walkers are greater than or equal to the pmin
        underweight_walker_idxs = np.where(walker_weights < self.pmin)[0]
        if len(underweight_walker_idxs) > 0:

            raise ResamplerError("All walker weights must be greater than the pmin, "
                                 "walkers {} are all underweight".format(
                                     ','.join([str(i) for i in underweight_walker_idxs])))



    def assign_clones(self, merge_groups, walker_clone_nums):
        """Convert two convenient data structures to a list of almost
        normalized resampling records.

        The two data structures are merge_groups and walker_clone_nums
        and are convenient to make.

        Each is a list with number of elements equal to the number of
        walkers that resampling will act on.

        Each element of the merge_groups is a list-like of integers
        indicating the indices of the walkers that will be merged into
        this one (i.e. squashed). A non-empty collection indicates a
        KEEP_MERGE decision.

        Each element of the walker_clone_nums is simply an integer
        specifying how many clones to make of this walker.

        These data structures simply declare requirements on what the
        actual decision records must achieve. The actual placement of
        walkers in slots (indices) is unspecified and immaterial.

        Parameters
        ----------
        merge_groups : list of list of int
            The specification of which walkers will be squashed and merged.

        walker_clone_nums : list of int
            The number of clones to make for each walker.

        Returns
        -------

        walker_actions : list of dict of str: values
            List of resampling record like dictionaries. These are not
            completely normalized for consumption by reporters, since
            they don't have the right list-like wrappers.

        """

        n_walkers = len(walker_clone_nums)

        walker_actions = self._init_walker_actions(n_walkers)

        # keep track of which slots will be free due to squashing
        free_slots = []

        # go through the merge groups and write the records for them,
        # the index of a merge group determines the KEEP_MERGE walker
        # and the indices in the merge group are the walkers that will
        # be squashed
        for walker_idx, merge_group in enumerate(merge_groups):

            if len(merge_group) > 0:
                # add the squashed walker idxs to the list of open
                # slots
                free_slots.extend(merge_group)

                # for each squashed walker write a record and save it
                # in the walker actions
                for squash_idx in merge_group:
                    walker_actions[squash_idx] = self.decision.record(self.decision.ENUM.SQUASH.value,
                                                                      target_idxs=(walker_idx,))

                # make the record for the keep merge walker
                walker_actions[walker_idx] = self.decision.record(self.decision.ENUM.KEEP_MERGE.value,
                                                         target_idxs=(walker_idx,))

        # for each walker, if it is to be cloned assign open slots for it
        for walker_idx, num_clones in enumerate(walker_clone_nums):

            if num_clones > 0 and len(merge_groups[walker_idx]) > 0:
                raise ResamplerError("Error! cloning and merging occuring with the same walker")

            # if this walker is to be cloned do so and consume the free
            # slot
            if num_clones > 0:

                # we first check to see if there are any free "slots"
                # for cloned walkers to go. If there are not we can
                # make room. The number of extra slots needed should
                # be default 0


                # we choose the targets for this cloning, we start
                # with the walker initiating the cloning
                clone_targets = [walker_idx]

                # if there are any free slots, then we use those first
                if len(free_slots) > 0:
                    clone_targets.extend([free_slots.pop() for clone in range(num_clones)])

                # if there are more slots needed then we will have to
                # create them
                num_slots_needed = num_clones - len(clone_targets)
                if num_slots_needed > 0:

                    # initialize the lists of open slots
                    new_slots = []

                    # and make a list of the new slots
                    new_slots = [n_walkers + i for i in range(num_slots_needed)]

                    # then increase the number of walkers to match
                    n_walkers += num_slots_needed

                    # then add these to the clone targets
                    clone_targets.extend(new_slots)

                # make a record for this clone
                walker_actions[walker_idx] = self.decision.record(self.decision.ENUM.CLONE.value,
                                                         target_idxs=tuple(clone_targets))

        return walker_actions



"""Reference implementations for a general walker and walker state
with utilities for cloning and merging walkers.

Wepy does not require that this or a subclass of this `Walker` or
`WalkerState` is used and only that it is something that acts like it
(duck typed).

The required attributes for the Walker interface are:

- `state` : object implementing the `WalkerState` interface
- `weight` : float

The weight attribute is simply a float and the proper normalization of
weights among a weighted ensemble of walkers should be enforced by the
resampler.

The `clone`, `squash`, and `merge` methods can be accessed by Decision
classes for implementing cloning and merging. As can the module level
`split` and `keep_merge` functions.

The WalkerState interface must have a method `dict` which returns a
dictionary with string keys and arbitrary values.

Additionally the WalkerState should provide its own `__getitem__`
magic method for the accessor syntax, i.e. walker.state['positions'].

"""

import random as rand
import logging
from copy import deepcopy

def split(walker, number=2):
    """Split (AKA make multiple clones) of a single walker.

    Creates multiple new walkers that have the same state as the given
    walker with weight evenly divided between them.

    Parameters
    ----------
    walker : object implementing the Walker interface
        The walker to split/clone
    number : int
        The number of clones to make of the walker
         (Default value = 2)

    Returns
    -------
    cloned_walkers : list of objects implementing the Walker interface

    """
    # calculate the weight of all child walkers split uniformly
    split_prob = walker.weight / (number)
    # make the clones
    clones = []
    for i in range(number):
        clones.append(type(walker)(walker.state, split_prob))

    return clones

def keep_merge(walkers, keep_idx):
    """Merge a set of walkers using the state of one of them.

    Parameters
    ----------
    walkers : list of objects implementing the Walker interface
        The walkers that will be merged together
    keep_idx : int
        The index of the walker in the walkers list that will be used
        to set the state of the new merged walker.

    Returns
    -------
    merged_walker : object implementing the Walker interface

    """

    weights = [walker.weight for walker in walkers]
    # but we add their weight to the new walker
    new_weight = sum(weights)
    # create a new walker with the keep_walker state
    new_walker = type(walkers[0])(walkers[keep_idx].state, new_weight)

    return new_walker

def merge(walkers):
    """Merge this walker with another keeping the state of one of them
    and adding the weights.

    The walker that has it's state kept is a random choice weighted by
    the walkers weights.

    Parameters
    ----------
    walkers : list of objects implementing the Walker interface
        The walkers that will be merged together

    Returns
    -------
    merged_walker : object implementing the Walker interface

    """

    weights = [walker.weight for walker in walkers]
    # choose a walker according to their weights to keep its state
    keep_walker = rand.choices(walkers, weights=weights)
    keep_idx = walkers.index(keep_walker)

    # TODO do we need this?
    # the others are "squashed" and we lose their state
    # squashed_walkers = set(walkers).difference(keep_walker)

    # but we add their weight to the new walker
    new_weight = sum(weights)
    # create a new walker with the keep_walker state
    new_walker = type(walkers[0])(keep_walker.state, new_weight)

    return new_walker, keep_idx

class Walker(object):
    """Reference implementation of the Walker interface.

    A container for:

    - state
    - weight

    """

    def __init__(self, state, weight):
        """Constructor for Walker.

        Parameters
        ----------
        state : object implementing the WalkerState interface

        weight : float

        """

        self.state = state
        self.weight = weight

    def clone(self, number=1):
        """Clone this walker by making a copy with the same state and split
        the probability uniformly between clones.

        The number is the increase in the number of walkers.

        e.g. number=1 will return 2 walkers with the same state as
        this object but with probability split 50/50 between them

        Parameters
        ----------
        number : int
            Number of extra clones to make
             (Default value = 1)

        Returns
        -------
        cloned_walkers : list of objects implementing the Walker interface

        """

        # calculate the weight of all child walkers split uniformly
        split_prob = self.weight / (number+1)
        # make the clones
        clones = []
        for i in range(number+1):
            clones.append(type(self)(self.state, split_prob))

        return clones

    def squash(self, merge_target):
        """Add the weight of this walker to another.

        Parameters
        ----------
        merge_target : object implementing the Walker interface
            The walker to add this one's weight to.

        Returns
        -------
        merged_walker : object implementing the Walker interface

        """
        new_weight = self.weight + merge_target.weight
        return type(self)(merge_target.state, new_weight)

    def merge(self, other_walkers):
        """Merge a set of other walkers into this one using the merge function.

        Parameters
        ----------
        other_walkers : list of objects implementing the Walker interface
            The walkers that will be merged together

        Returns
        -------
        merged_walker : object implementing the Walker interface

        """
        return merge([self]+other_walkers)

class WalkerState(object):
    """Reference implementation of the WalkerState interface.

    Access all key-value pairs as a dictionary with the dict() method.

    Access individual values using the accessor syntax similar to
    dictionaries:

    >>> WalkerState(my_key='value')['my_key']
    'value'

    """

    def __init__(self, **kwargs):
        """Constructor for WalkerState.

        All key-word arguments passed in will be set as the key-value
        pairs for the state.

        """
        self._data = kwargs

    def __getitem__(self, key):
        return self._data[key]

    def dict(self):
        """Return all key-value pairs as a dictionary."""
        return deepcopy(self._data)



"""This is a exmaple of random walke simultion using the REVO
resampler.
"""
import sys
import os
import os.path as osp
from pathlib import Path

import numpy as np

from wepy.resampling.resamplers.revo import REVOResampler

from wepy.resampling.distances.randomwalk import RandomWalkDistance
from wepy.runners.randomwalk import RandomWalkRunner, UNIT_NAMES
from wepy.walker import Walker, WalkerState
from wepy.work_mapper.mapper import Mapper

from wepy.sim_manager import Manager
from wepy.walker import Walker, WalkerState

from wepy.runners.randomwalk import RandomWalkRunner

from wepy_tools.sim_makers.toys.randomwalk import RandomwalkProfiler


ON = True
OFF = False
# the maximum weight allowed for a walker
PMAX = 0.1
# the minimum weight allowed for a walker
PMIN = 1e-100

# set the value of distance exponent
DIST_EXPONENT = 4
# the merge distance value
MERGE_DIST = 2.5
# field in the HDF5
SAVE_FIELDS = ('positions')
# Name of units of fields in the HDF5
UNITS = UNIT_NAMES
PROBABILITY=0.25
WEIGHTS=ON


outputs_dir = Path('_output')

if not osp.exists(outputs_dir):
    os.makedirs(outputs_dir)

# sets the input paths
hdf5_filename = 'rw_results.wepy.h5'
reporter_filename = 'randomwalk_revo.org'

hdf5_path= outputs_dir / hdf5_filename
reporter_path = outputs_dir / reporter_filename


def get_char_distance(dimension, num_walkers):
    """Calculate the characteristic value.
    Runs one cycle simulation and calculates the characteristic
    distance value.

    Parameters
    ----------
    dimension: int
        The dimension of the random walk space.


    num_walkers: int
        The number of walkers.

    Returns
    -------
    characteristic distance : float
        The characteristic distance value.

    """
    # set up initial state for walkers
    positions = np.zeros((1, dimension))

    init_state = WalkerState(positions=positions, time=0.0)

    # set up  the distance function
    rw_distance = RandomWalkDistance();

    # set up the  REVO Resampler with the parameters
    resampler = REVOResampler(distance=rw_distance,
                              pmin=PMIN, pmax=PMAX,
                              init_state=init_state,
                              char_dist=1,
                              merge_dist=MERGE_DIST
                              )

    # create list of init_walkers
    initial_weight = 1/num_walkers

    init_walkers = [Walker(init_state, initial_weight)
                    for i in range(num_walkers)]

    # set up raunner for system
    runner = RandomWalkRunner(probability=PROBABILITY)

    n_steps = 10
    mapper = Mapper()
    # running the simulation
    sim_manager = Manager(init_walkers,
                          runner=runner,
                          resampler=resampler,
                          work_mapper=mapper)


    print("Running simulation")
    #runs for one cycle
    sim_manager.init(num_walkers)

    new_walkers = sim_manager.run_segment(init_walkers, n_steps, 0)

    dist_matrix, _ = resampler._all_to_all_distance(new_walkers)


    return np.average(dist_matrix)



if __name__=="__main__":

    if sys.argv[1] == "--help" or sys.argv[1] == '-h':
        print("arguments: n_cycles, n_walkers, dimension")
    else:

        n_runs = int(sys.argv[1])
        n_cycles = int(sys.argv[2])
        n_walkers = int(sys.argv[3])
        dimension = int(sys.argv[4])


    dimension = 5

    # set up initial state for walkers
    position_coords = np.zeros((1, dimension))

    init_state = WalkerState(positions=position_coords, time=0.0)

    # set up  the distance function
    distance = RandomWalkDistance();


    char_dist = get_char_distance(dimension, n_walkers)
    # set up the Revo Resampler with the parameters
    resampler = REVOResampler(distance=distance,
                              pmin=PMIN,
                              pmax=PMAX,
                              dist_exponent=DIST_EXPONENT,
                              init_state=init_state,
                              char_dist=char_dist,
                              merge_dist=MERGE_DIST,
                              weights=WEIGHTS)


    # set up a RandomWalkProfilier
    rw_profiler = RandomwalkProfiler(resampler,
                                     dimension,
                                     hdf5_filename=str(hdf5_path),
                                     reporter_filename=str(reporter_path))

    # runs the simulations and gets the result
    rw_profiler.run(num_runs=n_runs, num_cycles=n_cycles,
                    num_walkers=n_walkers)



"""Compute rates for warping events."""

from collections import defaultdict

def calc_warp_rate(warping_records, total_sampling_time):
    """Calculate the warping rates for each target in these records.

    Takes a set of warping records which are named tuples that have
    the fields set in the Abstract Base Class for boundary conditions
    (wepy.boundary_conditions.boundary.Boundary).

    Namely these are: weight, target_idx, cycle_idx

    Parameters
    ----------

    warping_records : list of namedtuples implementing the warping interface
        The list of warping records for which events will be used to
        calculate the rates.

    total_sampling_time : float
        The total amount of sampling time that was run to generate
        these warping records.

    Returns
    -------

    target_weights_rates : dict of int : tuple of (float, float, float)
        A dictionary where each key is for a target present in the
        warping records and each value is a tuple giving the total
        accumulated weight in that target boundary, the total flux of
        probability across the boundary (rate), and the total sampling
        time it was calculated with,respectively.

    See Also
    --------

    wepy.boundary_conditions.boundary.Boundary : for specs on fields
                                                 necessary for warping_records

    """

    # we wish we could return a tuple of (0., np.inf) for no records
    # but we don't know the target indices so there isn't really a way
    # to do this without defaults which will only wreak havoc I am
    # sure. So this must be dealt with outside of this function.

    # get the weights from the records, the fields for the records are
    # defined in the base class for boundary conditions

    # we separate them based on the target
    target_warp_weights = defaultdict(list)
    for rec in warping_records:

        # get the weight
        target_warp_weights[rec.target_idx].append(rec.weight)

    # sum each target weights up, and compute the weights
    target_results = {}
    for target_key, weights in target_warp_weights.items():

        total_weight = sum(weights)
        rate = total_weight / total_sampling_time

        target_results[target_key] = (total_weight, rate, total_sampling_time)

    return target_results


def contig_warp_rates(contig, cycle_time, time_points=None):
    """Calculate the warping rates for a contig.

    Automatically figures out the total sampling time for two options
    which is set by the 'time_points' kwarg.

    If 'time_points' is None then a rate will only be computed for the
    end of the simulation.

    If 'time_points' is Ellipsis then a rate will be computed for each
    cycle in the entire run.

    Will work with runs that used boundary conditions deriving from
    and preserving the warping record fields from the abstract base
    class (wepy.boundary_conditions.boundary.Boundary).

    Parameters
    ----------

    contig : analysis.contig_tree.Contig
        Underlying WepyHDF5 must be open for reading.

    cycle_time : float
        The sampling time for each cycle.

    time_points : None or Ellipsis
        Determines how many and which time points are computed for the
        rates of each run. None, only the end; Ellipsis, all cycles.

    Returns
    -------

    run_target_weights_rates : list of dict of int : tuple of (float, float, float)

        List where each value of a run is a list of outputs from
        calc_warping_rate. The number and meaning of the elements of
        the list depend on the value of the 'time_points' kwarg.

    See Also
    --------

    wepy.analysis.rates.calc_warp_rate

    wepy.boundary_conditions.boundary.Boundary : for specs on fields
                                                 necessary for warping_records

    """

    # if time_points is None that means just get the rate for the
    # whole run or Ellipsis to evaluate it for all cycles.
    if time_points is None:
        n_cycle_points = [contig.num_cycles]
        cycle_idxs = [contig.num_cycles - 1]

    elif time_points is Ellipsis:
        n_cycle_points = [i + 1 for i in range(contig.num_cycles)]
        cycle_idxs = list(range(contig.num_cycles))

    # get a vector for the number of walkers for each cycle_idx
    n_walkers = [contig.num_walkers(cycle_idx) for cycle_idx in cycle_idxs]

    # calculate the total sampling time for each point which is the
    # cycle_index times the real sampling time for each walker in the
    # ensemble times the number of walkers
    total_sampling_times = [n_cycle * cycle_time * n_walkers[i]
                            for i, n_cycle in enumerate(n_cycle_points)]

    all_recs = contig.warping_records()

    contig_rates = []
    for cycle_idx, total_sampling_time in zip(cycle_idxs, total_sampling_times):
        recs = [rec for rec in all_recs if rec.cycle_idx <= cycle_idx]
        rate = calc_warp_rate(recs, total_sampling_time)
        contig_rates.append(rate)

    return contig_rates



import os
import os.path as osp
import pickle
import logging

class ReporterError(Exception):
    """ """
    pass

class Reporter(object):
    """Abstract base class for wepy reporters.

    All reporters must customize and override minimally the 'report'
    method. Optionally the 'init' and 'cleanup' can be overriden.

    See Also
    --------

    wepy.sim_manager : details of calls to reporter methods.

    """


    def __init__(self, **kwargs):
        """Construct a reporter.

        Void constructor for the Reporter base class.

        Parameters
        ----------
        **kwargs : key-value pairs
            Ignored kwargs, but accepts them from subclass calls for
            compatibility.

        """
        pass

    def init(self, **kwargs):
        """Initialization routines for the reporter at simulation runtime.

        Initialize I/O connections including file descriptors,
        database connections, timers, stdout/stderr etc.

        Void method for reporter base class.

        Reporters can expect to have the following key word arguments
        passed to them during a simulation by the sim_manager in this
        call.


        Parameters
        ----------

        init_walkers : list of Walker objects
            The initial walkers for the simulation.

        runner : Runner object
            The runner that will be used in the simulation.

        resampler : Resampler object
            The resampler that will be used in the simulation.

        boundary_conditions : BoundaryConditions object
            The boundary conditions taht will be used in the simulation.

        work_mapper : WorkMapper object
            The work mapper that will be used in the simulation.

        reporters : list of Reporter objects
            The list of reporters that are in the simulation.

        continue_run : int
            The index of the run that is being continued within this
            same file.

        """
        method_name = 'init'
        assert not hasattr(super(), method_name), \
            "Superclass with method {} is masked".format(method_name)

    def report(self, **kwargs):
        """Given data concerning the main simulation components state, perform
        I/O operations to persist that data.

        Void method for reporter base class.

        Reporters can expect to have the following key word arguments
        passed to them during a simulation by the sim_manager.

        Parameters
        ----------

        cycle_idx : int

        new_walkers : list of Walker objects
            List of walkers that were produced from running their
            dynamics by the runner.

        warp_data : list of dict of str : value
            List of dict-like records for each warping event from the
            last cycle.

        bc_data : list of dict of str : value
           List of dict-like records specifying the changes to the
           state of the boundary conditions in the last cycle.

        progress_data : dict str : list
            A record indicating the progress values for each walker in
            the last cycle.

        resampling_data : list of dict of str : value
            List of records specifying the resampling to occur at this
            cycle.

        resampler_data : list of dict of str : value
            List of records specifying the changes to the state of the
            resampler in the last cycle.

        n_segment_steps : int
            The number of dynamics steps that were completed in the last cycle

        worker_segment_times : dict of int : list of float
            Mapping worker index to the times they took for each
            segment they processed.

        cycle_runner_time : float
            Total time runner took in last cycle.

        cycle_bc_time : float
            Total time boundary conditions took in last cycle.

        cycle_resampling_time : float
            Total time resampler took in last cycle.

        resampled_walkers : list of Walker objects
            List of walkers that were produced from the new_walkers
            from applying resampling and boundary conditions.

        """

        method_name = 'report'
        assert not hasattr(super(), method_name), \
            "Superclass with method {} is masked".format(method_name)

    def cleanup(self, **kwargs):
        """Teardown routines for the reporter at the end of the simulation.

        Use to cleanly and safely close I/O connections or other
        cleanup I/O.

        Use to close file descriptors, database connections etc.

        Reporters can expect to have the following key word arguments
        passed to them during a simulation by the sim_manager.

        Parameters
        ----------

        runner : Runner object
            The runner at the end of the simulation

        work_mapper : WorkeMapper object
            The work mapper at the end of the simulation

        resampler : Resampler object
            The resampler at the end of the simulation

        boundary_conditions : BoundaryConditions object
            The boundary conditions at the end of the simulation

        reporters : list of Reporter objects
            The list of reporters at the end of the simulation

        """
        method_name = 'cleanup'
        assert not hasattr(super(), method_name), \
            "Superclass with method {} is masked".format(method_name)


class FileReporter(Reporter):
    """Abstract reporter that handles specifying file paths for a
    reporter.

    This abstract class doesn't perform any operations that involve
    actually opening file descriptors, but only the validation and
    organization of file paths.

    This provides a uniform API for retrieving file paths from all
    reporters inheriting from it.

    Additionally, FileReporter implements an interface for performing
    a so-called reparametrization of the relevant values associated
    with each file specification (i.e. file path and mode).

    A reparametrization can be performed by calling the
    'reparametrize' method, and can be customized.

    Additionally, there are some customizable class constants than can
    be used in subclasses to control this process including:
    DEFAULT_MODE, SUGGESTED_FILENAME_TEMPLATE,
    DEFAULT_SUGGESTED_EXTENSION, FILE_ORDER, and SUGGESTED_EXTENSIONS.

    The intention is to allow the redefinition of file paths
    dynamically to adapt to changing runtime requirements. Such as
    execution on a separate subtree of a directory hierarchy.

    """

    MODES = ('x', 'w', 'w-', 'r', 'r+',)
    """Valid modes accepted for files."""

    DEFAULT_MODE = 'x'
    """The default mode to set for opening files if none is specified
    (create if doesn't exist, fail if it does.)"""


    SUGGESTED_FILENAME_TEMPLATE = "{config}{narration}{reporter_class}.{ext}"
    """Template to use for dynamic reparametrization of file path names.

    The fields in the template are:

    config : indicator of the runtime configuration used

    narration : freeform description of the instance

    reporter_class : the name of the class that produced the
        output. When no specific name is given for a file report generated
        from a reporter this is used to disambiguate, along with the
        extension.

    ext : The file extension, for multiple files produced from one
    reporter this should be sufficient to disambiguate the files.

    The 'config' and 'narration' should be the same across all
    reporters in the same simulation manager, and the 'narration' is
    considered optional.

    """

    DEFAULT_SUGGESTED_EXTENSION = 'report'
    """The default file extension used for files during dynamic
    reparametrization, if none is specified"""

    FILE_ORDER = ()
    """Specify an ordering of file paths. Should be customized."""

    SUGGESTED_EXTENSIONS = ()
    """Suggested extensions for file paths for use with the automatic
    reparametrization feature. Should be customized."""

    def __init__(self, file_paths=None, modes=None,
                 file_path=None, mode=None,
                 **kwargs):
        """Constructor for FileReporter.

        This constructor allows the specification of either a list of
        file names (and modes) via 'file_paths' and 'modes' key-word
        arguments or a single 'file_path' and 'mode'.

        The access API though is always a list of file paths and modes
        where order is important for associating other features.

        Parameters
        ----------

        file_paths : list of str
            The list of file paths (in order) to use.

        modes : list of str
            The list of mode specs (in order) to use.

        file_path : str
            If 'file_paths' not specified, the single file path to use.

        mode : str
            If 'file_path' option used, this is the mode for that file.

        """

        # file paths

        assert not ((file_paths is not None) and (file_path is not None)), \
            "only file_paths or file_path kwargs can be specified"

        # if only one file path is given then we handle it as multiple
        if file_path is not None:
            file_paths = [file_path]


        # if any of the explicit paths are given (from the FILE_ORDER
        # constant) in the kwargs then we automatically add those to
        # the file_paths being sent to the super class constructor.

        # we use a flag to condition this, initialize and fall back to
        # using the 'file_paths' kwarg
        use_explicit_path_kwargs = False

        # we check the kwargs for the explicit file kwargs, and if
        # they are given then we check whether they are valid and if
        # they are, use them to set the 'file_paths' kwarg

        # make a list of the presence of the given explicit keys
        given_explicit_kwargs = [(True if file_key in kwargs else False)
                                for file_key in self.FILE_ORDER]

        # check that all the keys are present, if they aren't all
        # present then the flag will stay false and the fallback of
        # using the 'file_paths' kwarg will be used
        if all(given_explicit_kwargs):

            # then get the values and check them
            valid_explicit_kwargs = [(True if kwargs[file_key] is not None else False)
                                     for file_key in self.FILE_ORDER]


            # if they are all valid then we can use them
            if not all(valid_explicit_kwargs):

                use_explicit_path_kwargs = True

        # if only some were given this is wrong
        elif any(given_explicit_kwargs):

            raise ValueError("If you explicitly pass in the paths, all must be given explicitly")



        # if we use the explicit path kwargs, then we need to put them
        # into the 'file_paths' for superclass initialization
        if use_explicit_path_kwargs:

                file_paths = []
                for file_key in self.FILE_ORDER:

                    # add it to the file paths for superclass initialization
                    file_paths.append(kwargs[file_key])

        # otherwise we need to use the file_paths argument that should
        # have been given
        else:

            # make sure it is in kwargs and valid
            assert file_paths is not None, \
                "if no explicit file path is given the 'file_paths' must have a value"

            assert len(file_paths) == len(self.FILE_ORDER), \
                "you must give file_paths {} paths".format(len(self.FILE_ORDER))

        # using the file_path paths we got above we set them as
        # attributes in this object
        for i, file_key in enumerate(self.FILE_ORDER):
            setattr(self, file_key, file_paths[i])


        # set the underlying file paths
        self._file_paths = file_paths


        # modes

        assert not ((modes is not None) and (mode is not None)), \
            "only modes or mode kwargs can be specified"

        # if modes is None we make modes, from defaults if we have to
        if modes is None:

            # if mode is None set it to the default
            if modes is None and mode is None:
                mode = self.DEFAULT_MODE

            # if only one mode is given copy it for each file given
            modes = [mode for i in range(len(self._file_paths))]

        self._modes = modes

        super().__init__(**kwargs)


    def _validate_mode(self, mode):
        """Check if the mode spec is a valid one.

        Parameters
        ----------
        mode : str

        Returns
        -------
        valid : bool

        """
        if mode in self.MODES:
            return True
        else:
            return False

    @property
    def mode(self):
        """For single file path reporters the mode of that file."""
        if len(self._file_paths) > 1:
            raise ReporterError("there are multiple files and modes defined")

        return self._modes[0]

    @property
    def file_path(self):
        """For single file path reporters the file path to that file spec."""
        if len(self._file_paths) > 1:
            raise ReporterError("there are multiple files and modes defined")

        return self._file_paths[0]

    @property
    def file_paths(self):
        """The file paths for this reporter, in order."""
        return self._file_paths

    @file_paths.setter
    def file_paths(self, file_paths):
        """Setter for the file paths.

        Parameters
        ----------
        file_paths : list of str

        """
        for i, file_path in enumerate(file_paths):
            self.set_path(i, file_path)

    def set_path(self, file_idx, path):
        """Set the path for a single indexed file.

        Parameters
        ----------
        file_idx : int
            Index in the listing of files.
        path : str
            The new path to set for this file

        """
        self._paths[file_idx] = path

    @property
    def modes(self):
        """The modes for the files, in order."""
        return self._modes

    @modes.setter
    def modes(self, modes):
        """Setter for the modes.

        Parameters
        ----------
        modes : list of str

        """
        for i, mode in enumerate(modes):
            self.set_mode(i, mode)

    def set_mode(self, file_idx, mode):
        """Set the mode for a single indexed file.

        Parameters
        ----------
        file_idx : int
            Index in the listing of files.
        mode : str
            The new mode spec.

        """

        if self._validate_mode(mode):
            self._modes[file_idx] = mode
        else:
            raise ValueError("Incorrect mode {}".format(mode))


    def reparametrize(self, file_paths, modes):
        """Set the file paths and modes for all files in the reporter.

        Parameters
        ----------
        file_paths : list of str
            New file paths for each file, in order.
        modes : list of str
            New modes for each file, in order.

        """

        self.file_paths = file_paths
        self.modes = modes

class ProgressiveFileReporter(FileReporter):
    """Super class for a reporter that will successively overwrite the
    same file over and over again. The base FileReporter really only
    supports creation of file one time.

    """

    def init(self, **kwargs):
        """Construct a ProgressiveFileReporter.

        This is exactly the same as the FileReporter.


        Parameters
        ----------

        file_paths : list of str
            The list of file paths (in order) to use.

        modes : list of str
            The list of mode specs (in order) to use.

        file_path : str
            If 'file_paths' not specified, the single file path to use.

        mode : str
            If 'file_path' option used, this is the mode for that file.

        See Also
        --------

        wepy.reporter.reporter.FileReporter

        """

        super().init(**kwargs)

        # because we want to overwrite the file at every cycle we
        # need to change the modes to write with truncate. This allows
        # the file to first be opened in 'x' or 'w-' and check whether
        # the file already exists (say from another run), and warn the
        # user. However, once the file has been created for this run
        # we need to overwrite it many times forcefully.

        # go thourgh each file managed by this reporter
        for file_i, mode in enumerate(self.modes):

            # if the mode is 'x' or 'w-' we check to make sure the file
            # doesn't exist
            if mode in ['x', 'w-']:
                file_path = self.file_paths[file_i]
                if osp.exists(file_path):
                    raise FileExistsError("File exists: '{}'".format(file_path))

            # now that we have checked if the file exists we set it into
            # overwrite mode
            self.set_mode(file_i, 'w')



from collections import namedtuple, defaultdict
from enum import Enum
import logging

import numpy as np

from wepy.resampling.decisions.decision import Decision

from wepy.walker import split, keep_merge

# the possible types of decisions that can be made enumerated for
# storage, these each correspond to specific instruction type
class CloneMergeDecisionEnum(Enum):
    """Enum definition for cloning and merging decision values."

    - NOTHING : 1
    - CLONE : 2
    - SQUASH : 3
    - KEEP_MERGE : 4

    """

    NOTHING = 1
    """Do nothing with the walker sample. """

    CLONE = 2
    """Clone the walker into multiple children equally splitting the weight."""

    SQUASH = 3
    """Destroy the walker sample value (state) and donate the sample
    weight to another KEEP_MERGE walker sample."""

    KEEP_MERGE = 4
    """Do nothing with the sample value (state) but squashed walkers will
    donate their weight to it."""



class MultiCloneMergeDecision(Decision):
    """Decision encoding cloning and merging decisions for weighted ensemble.

    The decision records have in addition to the 'decision_id' a field
    called 'target_idxs'. This field has differing interpretations
    depending on the 'decision_id'.

    For NOTHING and KEEP_MERGE it indicates the walker index to assign
    this sample to after resampling. In this sense the walker is
    merely a vessel for the propagation of the state and acts as a
    slot.

    For SQUASH it indicates the walker that it's weight will be given
    to, which must have a KEEP_MERGE record for it.

    For CLONE it indicates the walker indices that clones of this one
    will be placed in. This field is variable length and the length
    corresponds to the number of clones.

    """


    ENUM = CloneMergeDecisionEnum

    DEFAULT_DECISION = ENUM.NOTHING

    FIELDS = Decision.FIELDS + ('target_idxs',)
    SHAPES = Decision.SHAPES + (Ellipsis,)
    DTYPES = Decision.DTYPES + (np.int,)

    RECORD_FIELDS = Decision.RECORD_FIELDS + ('target_idxs',)


    # the decision types that pass on their state
    ANCESTOR_DECISION_IDS = (ENUM.NOTHING.value,
                             ENUM.KEEP_MERGE.value,
                             ENUM.CLONE.value,)


    # TODO deprecate in favor of Decision implementation
    @classmethod
    def record(cls, enum_value, target_idxs):
        record = super().record(enum_value)
        record['target_idxs'] = target_idxs

        return record

    @classmethod
    def action(cls, walkers, decisions):

        # list for the modified walkers
        mod_walkers = [None for i in range(len(walkers))]

        # perform clones and merges for each step of resampling
        for step_idx, step_recs in enumerate(decisions):
            # we need to collect groups of merges, one entry for each
            # merge, where the key is the walker_idx of the keep merge slot
            squash_walkers = defaultdict(list)
            keep_walkers = {}
            # go through each decision and perform the decision
            # instructions
            for walker_idx, walker_rec in enumerate(step_recs):

                decision_value = walker_rec['decision_id']
                instruction = walker_rec['target_idxs']

                if decision_value == cls.ENUM.NOTHING.value:
                    # check to make sure a walker doesn't already exist
                    # where you are going to put it
                    if mod_walkers[instruction[0]] is not None:
                        raise ValueError(
                            "Multiple walkers assigned to position {}".format(instruction[0]))

                    # put the walker in the position specified by the
                    # instruction
                    mod_walkers[instruction[0]] = walkers[walker_idx]

                # for a clone
                elif decision_value == cls.ENUM.CLONE.value:

                    # get the walker to be cloned
                    walker = walkers[walker_idx]
                    # "clone" it by splitting it into walkers of the
                    # same state with even weights
                    clones = split(walker, number=len(instruction))

                    # then assign each of these clones to a target
                    # walker index in the next step
                    for clone_idx, target_idx in enumerate(instruction):

                        # check that there are not another walker
                        # already assigned to this position
                        if mod_walkers[target_idx] is not None:
                            raise ValueError(
                                "Multiple walkers assigned to position {}".format(instruction[0]))

                        # TODO this comment was just fixed so I
                        # believe that there was some serious problems
                        # before
                        # mod_walkers[walker_idx] = clones[clone_idx]

                        # assign the clone to the modified walkers of the next step
                        mod_walkers[target_idx] = clones[clone_idx]

                # if it is a decision for merging we must perform this
                # once we know all the merge targets for each merge group
                elif decision_value == cls.ENUM.SQUASH.value:

                    # save this walker to the appropriate merge group to
                    # merge after going through the list of walkers
                    squash_walkers[instruction[0]].append(walker_idx)

                elif decision_value == cls.ENUM.KEEP_MERGE.value:
                    keep_walkers[instruction[0]] = walker_idx

                else:
                    raise ValueError("Decision not recognized")

            # do the merging for each merge group
            for target_idx, walker_idxs in squash_walkers.items():

                keep_idx = keep_walkers[target_idx]

                # collect the walkers in the merge group, the keep idx is
                # always the first in the list
                merge_grp = [walkers[keep_idx]] + [walkers[i] for i in walker_idxs]

                # merge the walkers
                merged_walker = keep_merge(merge_grp, 0)

                # make sure there is not already a walker in this slot
                if mod_walkers[target_idx] is not None:
                    raise ValueError(
                        "Multiple walkers assigned to position {}".format(target_idx))

                # set it in the slot for the keep_idx
                mod_walkers[keep_idx] = merged_walker


        if not all([False if walker is None else True for walker in mod_walkers]):

            raise ValueError("Some walkers were not created")

        return mod_walkers



import json

import pandas as pd

CHAIN_KEY = 'index'
RESIDUE_KEY = 'index'
ATOM_KEY = 'index'
CHAIN_ATTR_KEYS = (CHAIN_KEY,)
RESIDUE_ATTR_KEYS = (RESIDUE_KEY, 'name', 'resSeq', 'segmentID')
ATOM_ATTR_KEYS = (ATOM_KEY, 'name', 'element')

def json_top_chain_fields(json_topology):
    """Creates a fields dictionary for all of the chains in a topology.

    Parameters
    ----------

    json_topology : str
        JSON format topology

    Returns
    -------

    chain_cols : dict of str: list

    """


    top = json.loads(json_topology)

    chain_cols = {key : [] for key in CHAIN_ATTR_KEYS}

    for chain in top['chains']:
        for key in CHAIN_ATTR_KEYS:
            chain_cols[key].append(chain[key])

    return chain_cols

def json_top_chain_df(json_topology):
    """Make a pandas dataframe for the chains in the JSON topology.

    Parameters
    ----------

    json_topology : str
        JSON format topology

    Returns
    -------

    chain_df : pandas.DataFrame

    """

    return pd.DataFrame(json_top_chain_fields(json_topology))

def json_top_residue_fields(json_topology):
    """Creates a fields dictionary for all of the residues in a topology.

    Parameters
    ----------

    json_topology : str
        JSON format topology

    Returns
    -------

    residue_cols : dict of str: list

    """


    top = json.loads(json_topology)

    residue_cols = {key : [] for key in RESIDUE_ATTR_KEYS}
    residue_cols['chain_key'] = []

    for chain in top['chains']:
        for residue in chain['residues']:
            residue_cols['chain_key'].append(chain[CHAIN_KEY])
            for key in RESIDUE_ATTR_KEYS:
                residue_cols[key].append(residue[key])

    return residue_cols

def json_top_residue_df(json_topology):
    """Make a pandas dataframe for the residues in the JSON topology.

    Parameters
    ----------

    json_topology : str
        JSON format topology

    Returns
    -------

    residue_df : pandas.DataFrame

    """

    return pd.DataFrame(json_top_residue_fields(json_topology))


def json_top_atom_fields(json_topology):
    """Creates a fields dictionary for all of the atoms in a topology.

    Parameters
    ----------

    json_topology : str
        JSON format topology

    Returns
    -------

    atom_cols : dict of str: list

    """

    top = json.loads(json_topology)

    atom_cols = {key : [] for key in ATOM_ATTR_KEYS}
    atom_cols['chain_key'] = []
    atom_cols['residue_key'] = []

    for chain in top['chains']:
        for residue in chain['residues']:
            for atom in residue['atoms']:
                atom_cols['chain_key'].append(chain[CHAIN_KEY])
                atom_cols['residue_key'].append(residue[RESIDUE_KEY])
                for key in ATOM_ATTR_KEYS:
                    atom_cols[key].append(atom[key])

    return atom_cols

def json_top_atom_df(json_topology):
    """Make a pandas dataframe for the atoms in the JSON topology.

    Parameters
    ----------

    json_topology : str
        JSON format topology

    Returns
    -------

    atoms_df : pandas.DataFrame

    """

    return pd.DataFrame(json_top_atom_fields(json_topology))


def json_top_atom_count(json_str):
    """Count the number of atoms in the JSON topology used by wepy HDF5.

    Parameters
    ----------
    json_str : str
        A string of valid JSON in the format of JSON used in WepyHDF5
        and mdtraj HDF5 format.

    Returns
    -------
    n_atoms : int
        Number of atoms in the topology.

    """

    top_d = json.loads(json_str)
    atom_count = 0
    atom_count = 0
    for chain in top_d['chains']:
        for residue in chain['residues']:
            atom_count += len(residue['atoms'])

    return atom_count

def json_top_subset(json_str, atom_idxs):
    """Given a JSON topology and atom indices from that topology returns
    another JSON topology which is a subset of the first, preserving
    the topology between remaining atoms. The atoms will be ordered in
    the order in which the indices are given.

    Parameters
    ----------

    json_str : str
        A string of valid JSON in the format of JSON used in WepyHDF5
        and mdtraj HDF5 format.

    atom_idxs : list of int
        The atoms for which you want to make a subset of.

    Returns
    -------

    subset_json_str : str
        JSON string of the subset of atoms. Ordering preserved.

    """

    # cast so we can use the list index method
    atom_idxs = list(atom_idxs)

    # do checks on the atom idxs

    # no duplicates
    assert len(set(atom_idxs)) == len(atom_idxs), "duplicate atom indices"

    top = json.loads(json_str)

    # the dictionaries for each thing indexed by their old index
    atom_data_ds = {}
    residue_data_ds = {}

    # mapping of old atom indices to the residue data they belong to
    atom_res_idxs = {}
    res_chain_idxs = {}

    # go through and collect data on the atoms and convert indices to
    # the new ones
    for chain in top['chains']:

        for residue in chain['residues']:
            res_chain_idxs[residue['index']] = chain['index']

            # create the data dict for this residue
            residue_data_ds[residue['index']] = residue



            for atom in residue['atoms']:
                atom_res_idxs[atom['index']] = residue['index']

                # if the current atom's index is in the selection
                if atom['index'] in atom_idxs:

                    # we add this to the mapping by getting the index
                    # of the atom in subset
                    new_idx = atom_idxs.index(atom['index'])

                    # save the atom attributes
                    atom_data_ds[atom['index']] = atom

    # initialize the data structure for the topology subset
    top_subset = {'chains' : [],
                  'bonds' : []}

    residue_idx_map = {}
    chain_idx_map = {}

    old_to_new_atoms = {}

    # initialize the new indexing of the chains and residues
    new_res_idx_counter = 0
    new_chain_idx_counter = 0
    # now in the new order go through and create the topology
    for new_atom_idx, old_atom_idx in enumerate(atom_idxs):

        old_to_new_atoms[old_atom_idx] = new_atom_idx

        atom_data = atom_data_ds[old_atom_idx]

        # get the old residue index
        old_res_idx = atom_res_idxs[old_atom_idx]

        # since the datastrucutre is hierarchical starting with the
        # chains and residues we work our way back up craeting these
        # if neceessary for this atom, once this is taken care of we
        # will add the atom to the data structure
        if old_res_idx not in residue_idx_map:
            residue_idx_map[old_res_idx] = new_res_idx_counter
            new_res_idx_counter += 1

            # do the same but for the chain
            old_chain_idx = res_chain_idxs[old_res_idx]

            # make it if necessary
            if old_chain_idx not in chain_idx_map:
                chain_idx_map[old_chain_idx] = new_chain_idx_counter
                new_chain_idx_counter += 1

                # and add the chain to the topology
                new_chain_idx = chain_idx_map[old_chain_idx]
                top_subset['chains'].append({'index' : new_chain_idx,
                                             'residues' : []})

            # add the new index to the dats dict for the residue
            res_data = residue_data_ds[old_res_idx]
            res_data['index'] = residue_idx_map[old_res_idx]
            # clear the atoms
            res_data['atoms'] = []

            # add the reside to the chain idx
            new_chain_idx = chain_idx_map[old_chain_idx]
            top_subset['chains'][new_chain_idx]['residues'].append(res_data)

        # now that (if) we have made the necessary chains and residues
        # for this atom we replace the atom index with the new index
        # and add it to the residue
        new_res_idx = residue_idx_map[old_res_idx]
        new_chain_idx = chain_idx_map[res_chain_idxs[old_res_idx]]

        atom_data['index'] = new_atom_idx

        top_subset['chains'][new_chain_idx]['residues'][new_res_idx]['atoms'].append(atom_data)

    # then translate the atom indices in the bonds
    new_bonds = []
    for bond_atom_idxs in top['bonds']:

        if all([True if a_idx in old_to_new_atoms else False
                for a_idx in bond_atom_idxs]):
            new_bond_atom_idxs = [old_to_new_atoms[a_idx] for a_idx in bond_atom_idxs
                                  if a_idx in old_to_new_atoms]
            new_bonds.append(new_bond_atom_idxs)

    top_subset['bonds'] = new_bonds


    return json.dumps(top_subset)



"""The random walk dynamics runner.

In this system, the state of the walkers is defined as an
N-dimensional vector of non-negative values. The walkers start at
position zero (in N-dimensional space) and randomly move a step either
forward or backward with the given probabilities. This is done in each
dimension at each dynamic step. All moves that result in a negative
position are rejected.

One potentioanl use of the random walk system is to test the
performance of differnt resamplers as seen in these papers:

"WExplore: Hierarchical Exploration of High-Dimensional Spaces
Using the Weighted Ensemble Algorithm" and
"REVO: Resampling of Ensembles by Variation Optimization".

"""

import random as rand
import logging

import numpy as np
from pint import UnitRegistry

from wepy.runners.runner import Runner
from wepy.walker import Walker, WalkerState

units = UnitRegistry()

# the names of the units. We pass them through pint just to validate
# them
UNIT_NAMES = (('positions_unit', str(units('microsecond').units)),
         ('time_unit', str(units('picosecond').units)),
        )

"""Mapping of units identifiers to the corresponding pint units."""


class RandomWalkRunner(Runner):
    """RandomWalk runner for random walk simulations."""

    def __init__(self, probability=0.25):
        """Constructor for RandomWalkRunner.

        Parameters
        ----------

        probabilty : float
            "Probability" is defined here as the forward-move
             probability only. The backward-move probability is
             1-probability.(Default = 0.25)

        """

        self._probability = probability


    @property
    def probability(self):
        """ The probability of forward-move in an N-dimensional space"""
        return self._probability


    def _walk(self, positions):
        """Run dynamics for the RandomWalk system for one step.

        Parameters
        ----------
        positions : arraylike of shape (1, dimension)
            Current position of the walker.

        Returns
        -------
        new_positions : arraylike of shape (1, dimension)
            The positions of the walker after one dynamic step.

        """

        # make the deep copy of current posiotion
        new_positions = positions.copy()

        # get the dimension of the random walk space
        dimension = new_positions.shape[1]

        # iterates over each dimension
        for dim_idx in range(dimension):
            # Generates an uniform random number to choose between
            # moving forward or backward.
            rand_num = rand.uniform(0, 1)

            # make a forward movement
            if rand_num < self.probability:
                new_positions[0][dim_idx] += 1
            # make a backward movement
            else:
                new_positions[0][dim_idx] -= 1

            # implement the boundary condition for movement, movements
            # to -1 are rejected
            if new_positions[0][dim_idx] < 0:
                new_positions[0][dim_idx] = 0

        return new_positions

    def run_segment(self, walker, segment_length,
                    **kwargs):
        """Runs a random walk simulation for the given number of steps.

        Parameters
        ----------
        walker : object implementing the Walker interface
            The walker for which dynamics will be propagated.


        segment_length : int
            The numerical value that specifies how much dynamical steps
            are to be run.

        Returns
        -------
        new_walker : object implementing the Walker interface
            Walker after dynamics was run, only the state should be modified.

        """

        # Gets the current posiotion of RandomWalk Walker
        positions = walker.state['positions']

        # Make movements for the segment_length steps
        for _ in range(segment_length):
            # calls walk function for one step movement
            new_positions = self._walk(positions)
            positions = new_positions

        # makes new state form new positions
        new_state = WalkerState(positions=new_positions, time=0.0)

        # creates new_walker from new state and current weight
        new_walker = Walker(new_state, walker.weight)

        return new_walker



"""This is a exmaple of random walke simultion using the WExplore
resampler.

"""
import sys
import os
import os.path as osp
from pathlib import Path

import numpy as np

from wepy.resampling.resamplers.wexplore import WExploreResampler
from wepy.resampling.distances.randomwalk import RandomWalkDistance
from wepy.runners.randomwalk import RandomWalkRunner, UNIT_NAMES
from wepy.walker import Walker, WalkerState

from wepy_tools.sim_makers.toys.randomwalk import RandomwalkProfiler


ON = True
OFF = False
# the maximum weight allowed for a walker
PMAX = 0.1
# the minimum weight allowed for a walker
PMIN = 1e-100

# set the value of distance exponent
DIST_EXPONENT = 4
# the merge distance value
MERGE_DIST = 2.5
# field in the HDF5
SAVE_FIELDS = ('positions')
# Name of field's unit in the HDF5
UNITS = UNIT_NAMES
PROBABILITY=0.25
WEIGHTS=ON

# the maximum number of regions allowed under each parent region
MAX_N_REGIONS = (10, 10, 10, 10)

# the maximum size of regions, new regions will be created if a walker
# is beyond this distance from each voronoi image unless there is an
# already maximal number of regions
MAX_REGION_SIZES = (16, 4, 1, .25)

outputs_dir = Path('_output')
if not osp.exists(outputs_dir):
    os.makedirs(outputs_dir)


# sets the input paths
hdf5_filename = 'rw_results.wepy.h5'
reporter_filename = 'randomwalk_wexplore.org'

hdf5_path= outputs_dir / hdf5_filename
reporter_path = outputs_dir / reporter_filename



if __name__=="__main__":
    if sys.argv[1] == "--help" or sys.argv[1] == '-h':
        print("arguments: n_cycles, n_walkers, dimension")
    else:

        n_runs = int(sys.argv[1])
        n_cycles = int(sys.argv[2])
        n_walkers = int(sys.argv[3])
        dimension = int(sys.argv[4])


    dimension = 5

    # set up initial state for walkers
    position_coords = np.zeros((1, dimension))

    init_state = WalkerState(positions=position_coords, time=0.0)

    # set up  the distance function
    distance = RandomWalkDistance();



    # set up the WExplore Resampler with the parameters
    resampler = WExploreResampler(distance=distance,
                                  init_state=init_state,
                                  max_n_regions=MAX_N_REGIONS,
                                  max_region_sizes=MAX_REGION_SIZES,
                                  pmin=PMIN, pmax=PMAX)


    # set up a RandomWalkProfilier
    rw_profiler = RandomwalkProfiler(resampler,
                                     dimension,
                                     hdf5_filename=str(hdf5_path),
                                     reporter_filename=str(reporter_path))

    # runs the simulations and gets the result
    rw_profiler.run(num_runs=n_runs, num_cycles=n_cycles,
                    num_walkers=n_walkers)

        #set up the Wexplore Resampler with the parameters



"""Resampling functionality for weighted ensemble simulations.

Role of Resampling in WE
------------------------

The defining feature of weighted ensemble simulations (as opposed to
simply running multiple independent simulations in parallel) is the
addition of a processing step that allows for communication of the
states of the parallel simulations. In this communication step we have
the opportunity to make decisions about the relative productivity of
each of them and attempt to bias the overall behavior of the ensemble
of simulations, while not explicitly modifying the governing laws
(e.g. equations of motion, force field, hamiltonian, etc.) of each
parallel dynamics simulation universe.

A resampler embodies this communication. At a basic level it takes as
input a set of samples derived from dynamics, called walkers. In
addition to the states (the random variable) the walkers also have an
associated probability or weight. At the beginning of a WE simulation
each walker in the ensemble has a uniform probability. But the
probabilities are free to change due to the action of the resampler,
that is due to a resampling process.

Resampling characterizes a potentially large number of algorithms that
may or may not be useful. In the original WE algorithm by Huber and
Kim (TODO add citation) this resampling is achieved through the
"cloning" and "merging" of walkers.

Outline of the resampler sub-package
------------------------------------

The design of the wepy framework allows for any user to implement
their own resampler.

The only strict requirement that the framework is that a resampler
must have a method "resample" that receives a list of walkers and
returns some other list of walkers along with optional records the
resampling process and any changes to the state of the resampler
itself (while these records aren't strictly necessary for a simulation
to run, they are paramount to actually recording data on the
simulations).

However, useful resamplers share a lot of shared functionality. This
sub-package attempts to provide a sub-framework that provides this
functionality.

Decisions
=========

First is the definition of the "decisions" and the decision records.

A resampling process given a set of weighted samples should produce a
number of other samples where the set of states in the output is a
subset of those in the input.

For example a simple form of resampling is subsampling, where we
simply drop some of the samples from the collection. Similar to how
outliers are removed from a dataset.

In the subsampling mechanism we might say that there are only two
decisions a resampler can make: DROP and KEEP.

These could then be embodied in a Decision class
(e.g. SubsamplingDecision) which enumerates the possibilities and
provides methods for performing the operations on a collection of
walkers given a set of these decisions.

Such a resampler wouldn't be very useful for WE since the simulation
would simply fade away as each walker was eventually eliminated.

A more complex (and useful) example would be to add a possible
decision to split up a walker into two. In this there are now two
samples of the same random variable (state) but the total probability
of that state is the same since the weights are normalized and divided
in half for each.

These decisions might be called: DROP, KEEP, SPLIT. Now we can replace
dropped samples so that the number of samples is kept the same.

We could also allow the splitting of a walker into multiple walkers
instead of just two, in this case we replace SPLIT with CLONE. The
CLONE record now needs some additional information to parametrize it's
action. A record containing the decision value and the decision
parameters is called a 'decision record'. A CLONE decision record would
look like for example:

(decision=CLONE, n_children=3)

If we add data about which walker it applies to on which cycle and
which step of resampling it would be a 'resampling record'.


One issue with the use of a pure subsampling scheme is that we must
renormalize the weights of the walkers when any are dropped. In WE the
sample distribution is preserved at each step by allowing the
'merging' of walkers instead of just removing them. Merging involves
choosing a walker/sample to remove and adding that weight to another
walker, before discarding the sample value.

This cloning-merging strategy is the primary Decision class used in
wepy and the choice of most resamplers. The full listing of decisions
is:

- NOTHING
- CLONE
- SQUASH
- KEEP_MERGE

Where NOTHING and CLONE are as above, SQUASH is the walker that will
be discarded and have it's weight donated, and KEEP_MERGE is a walker
that will be receiving weight from one or more squashed walkers.

Resamplers
==========

Abstractly a resampler in the context of wepy is an algorithm that
given a sample of walkers produces instructions on how to resample
them. The instruction set is defined and implemented by a decision
object.

Concretely (and state above), the Resampler class is parametrized by a
Decision class and has a method 'resample' which receives a set of
walkers and returns a resampled collection of walkers.

Resamplers in the sub-framework also produce resampling records in
roughly the format described above so that storage formats can record
the lineages (family trees) of walkers.

Changes in the state of the resampler can also be recorded for
profiling and analysis purposes if it is desired, but is completely
specific to the resampling algorithm.


It is important to note that the choice of resampler is entirely
independent of:

- the type of sampling used e.g. molecular dynamics, monte carlo
- the system being simulated e.g. proteins, polymers

The type of sampling being used is encapsulated in the Runner class.

For the second point we would like our resamplers to be as resuable as
possible between different systems.

For example, wepy has implementations of both the WExplore and REVO
resampling algorithms which are broadly applicable and interchangable.

For instance if you want to simulate ligand unbinding for a particular
protein-ligand system you can choose either WExplore or REVO. This
choice can be made after you have already determined a metric for
discerning novelty in you protein-ligand system, as long as the metric
is encapsulated in a Distance class.

Distances
=========

A distance metric (in the context of wepy) is simply a function that
returns a single value when given the states of two walkers.

For such a function to be considered a proper distance metric extra
criterion are required but wepy does not enforce these in
implementation.

A Distance class is simply a class with a single method 'distance'. To
allow for certain basic performance requirements, it is also required
that the 'image' and 'image_distance' methods be overriden. This
allows for the transformation (which is potentially compute intensive)
of a walker state to a reduced form called an image. The
'image_distance' method should produce the same results as the
'distance' method for the same walker states.

"""



from copy import copy

from wepy_tools.sim_makers.openmm import OpenMMToolsTestSysSimMaker

from wepy_tools.systems import receptor as receptor_tools

from wepy.runners.openmm import GET_STATE_KWARG_DEFAULTS
from wepy.resampling.distances.receptor import UnbindingDistance
from wepy.boundary_conditions.receptor import UnbindingBC

from wepy.util.json_top import (json_top_residue_fields,
                                json_top_residue_df,
                                json_top_atom_df,
                                json_top_subset)

from wepy.reporter.receptor.dashboard import (
    UnbindingBCDashboardSection,
)


import mdtraj as mdj
import numpy as np
import simtk.unit as unit

from openmm_systems.test_systems import LysozymeImplicit

class LysozymeImplicitOpenMMSimMaker(OpenMMToolsTestSysSimMaker):

    TEST_SYS = LysozymeImplicit

    LIGAND_RESNAME = 'TMP'
    RECEPTOR_RES_IDXS = list(range(162))

    GET_STATE_KWARGS = {'enforcePeriodicBox' : False}

    BCS = OpenMMToolsTestSysSimMaker.BCS + [UnbindingBC]

    UNBINDING_BC_DEFAULTS = {
        'cutoff_distance' : 1.0, # nm
        'periodic' : False,
    }

    DEFAULT_BC_PARAMS = OpenMMToolsTestSysSimMaker.DEFAULT_BC_PARAMS
    DEFAULT_BC_PARAMS.update(
        {
            'UnbindingBC' : UNBINDING_BC_DEFAULTS,
        }
    )

    BOX_SIDE_LENGTH = 20 * unit.nanometer

    WEXPLORE_DEFAULTS = {
        'pmax' : 0.5,
        'pmin' : 1e-12,
        'max_n_regions' : (10, 10, 10, 10),
        'max_region_sizes' : (1, 0.5, 0.35, 0.25),
    }


    REVO_DEFAULTS = {
        'pmax' : 0.5,
        'pmin' : 1e-12,
        'char_dist' : 1,
        'merge_dist' : 0.25,
        'dist_exponent' : 4,
        'weights' : True,
    }


    DEFAULT_RESAMPLER_PARAMS = OpenMMToolsTestSysSimMaker.DEFAULT_RESAMPLER_PARAMS
    DEFAULT_RESAMPLER_PARAMS.update(
        {
            'WExploreResampler' : WEXPLORE_DEFAULTS,
            'REVOResampler' : REVO_DEFAULTS,
        }
    )

    def __init__(self, bs_cutoff=0.8*unit.nanometer):

        # must set this here since we need it to generate the state,
        # will get called again in the superclass method
        self.getState_kwargs = dict(GET_STATE_KWARG_DEFAULTS)
        if self.GET_STATE_KWARGS is not None:
            self.getState_kwargs.update(self.GET_STATE_KWARGS)

        test_sys = LysozymeImplicit()

        # set the box vectors to something reasonable

        # we have a constructor which uses a constant we set in the
        # class to do this
        box_vectors = self.box_vectors()

        # set them to the test system
        test_sys.system.setDefaultPeriodicBoxVectors(
            box_vectors[0],
            box_vectors[1],
            box_vectors[2],
        )

        init_state = self.make_state(test_sys.system, test_sys.positions)

        lig_idxs = self.ligand_idxs()
        bs_idxs = self.binding_site_idxs(bs_cutoff)

        distance = UnbindingDistance(ligand_idxs=lig_idxs,
                                     binding_site_idxs=bs_idxs,
                                     ref_state=init_state)


        super().__init__(
            distance=distance,
            init_state=init_state,
            system=test_sys.system,
            topology=test_sys.topology,
        )

    @classmethod
    def box_vectors(cls):

        # the box vectors from the test system are garbage and don't
        # matter for the simulation, however they cause lots of
        # problems when they are set to be super small. You should be
        # running things such that there is no periodic boundary
        # conditions but this is annoying and error-prone so we set
        # them to something large here just to make things DWIM more

        bvs = np.zeros((3,3))

        for dim in [0,1,2]:
            bvs[dim,dim] = cls.BOX_SIDE_LENGTH.value_in_unit(unit.nanometer)

        bvs = bvs * unit.nanometer

        return bvs

    def make_apparatus(self, **kwargs):

        # just customize an option to the runner to not enforce
        # periodic box
        runner_params = {'enforce_box' : False}

        resampler_params = None

        apparatus = super().make_apparatus(runner_params=runner_params,
                                           **kwargs)

        return apparatus

    @classmethod
    def ligand_idxs(cls):

        json_top = cls.json_top()

        res_df = json_top_residue_df(json_top)
        residue_idxs = res_df[res_df['name'] == cls.LIGAND_RESNAME]['index'].values


        # get the atom dataframe and select them from the ligand residue
        atom_df = json_top_atom_df(json_top)
        atom_idxs = atom_df[atom_df['residue_key'].isin(residue_idxs)]['index'].values

        return atom_idxs

    @classmethod
    def receptor_idxs(cls):

        json_top = cls.json_top()

        # get the atom dataframe and select them from the ligand residue
        atom_df = json_top_atom_df(json_top)
        atom_idxs = atom_df[atom_df['residue_key'].isin(cls.RECEPTOR_RES_IDXS)]['index'].values

        return atom_idxs

    @classmethod
    def binding_site_idxs(cls, cutoff):
        """Find the atoms that are within a distance of the ligand on the
        receptor.

        Parameters
        ----------

        cutoff : Quantity

        """

        test_sys = LysozymeImplicit()

        json_top = cls.json_top()

        atom_idxs = receptor_tools.binding_site_idxs(
            json_top,
            cls.ligand_idxs(),
            cls.receptor_idxs(),
            test_sys.positions,
            cls.box_vectors(),
            cutoff,
            periodic=False
        )

        return atom_idxs

    def choose_dashboard_sections(self, apparatus):

        dashboard_sections = super().choose_dashboard_sections(apparatus)

        if type(apparatus.boundary_conditions).__name__ == 'UnbindingBC':
            dashboard_sections['bc'] = UnbindingBCDashboardSection(apparatus.boundary_conditions)

        return dashboard_sections

    def make_bc(self, bc_class, bc_params):

        if bc_class == UnbindingBC:
            bc_params.update(
                {
                    'distance' : self.distance,
                    'initial_state' : self.init_state,
                    'topology' : self.json_top(),
                    'ligand_idxs' : self.ligand_idxs(),
                    'receptor_idxs' : self.receptor_idxs(),
                }
            )

        bc = bc_class(**bc_params)

        return bc



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import setup, find_packages

import itertools as it

import versioneer

# setuptools only specifies abstract requirements. For the concrete
# requirements i.e. index or repo URL see requirements.txt
base_requirements = [
    'numpy',
    'h5py',
    'networkx==2.3',
    'pandas',
    'dill',
    'click',
    'scipy',
    'geomm',
    'matplotlib',
    'tabulate',
    'jinja2',
    'pint',
    'multiprocessing_logging',
    'eliot',
]

# extras requirements list
md_requirements = [
    'mdtraj',
    'openmm_systems==0.0.0',
]
distributed_requirements = ['dask[bag]']

prometheus_requirements = [
    'prometheus_client',
    'pympler',
]

# # combination of all the extras requirements
all_requirements = list(it.chain.from_iterable([
    base_requirements,
    md_requirements,
    distributed_requirements,
    prometheus_requirements,
]))

setup(
    name='wepy',
    version=versioneer.get_version(),
    author="Samuel D. Lotz",
    author_email="samuel.lotz@salotz.info",
    description="Weighted Ensemble Framework",
    #long_description=open('README.org').read(),
    license="MIT",
    url="https://github.com/ADicksonLab/wepy",
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3'
    ],
    # building/dev
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'tox'],

    cmdclass=versioneer.get_cmdclass(),

    # package
    packages=find_packages(where='src'),

    package_dir={'' : 'src'},

    # if this is true then the package_data won't be included in the
    # dist. Use MANIFEST.in for this
    include_package_data=True,

    # pymodules is for single file standalone modules not part of the
    # package
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],

    entry_points={
        'console_scripts' : [
            'wepy=wepy.cli:cli',
        ],
        'pytest11' : [
            'pytest-wepy=pytest_wepy',
        ],

    },

    install_requires=base_requirements,

    extras_require={
        'md' : md_requirements,
        'distributed' : distributed_requirements,
        'prometheus' : prometheus_requirements,
        'all' : all_requirements,
    }
)



"""Miscellaneous functions needed by wepy."""

import json
import warnings

import numpy as np



def traj_box_vectors_to_lengths_angles(traj_box_vectors):
    """Convert box vectors for multiple 'frames' (a 'trajectory') to box lengths and angles.

    Parameters
    ----------
    traj_box_vectors : arraylike of float of shape (n_frames, n_dims, n_dims)
        Box vector matrices (x, y, z) for a trajectory of frames.

    Returns
    -------
    traj_box_lengths : arraylike of float of shape (n_frames, n_dims)
        The lengths of the box for each frame

    traj_box_angles : arraylike of float of shape (n_frames, n_dims)
        The angles of the box vectors to normal for each frame in degrees.

    """

    traj_unitcell_lengths = []
    for basis in traj_box_vectors:
        traj_unitcell_lengths.append(np.array([np.linalg.norm(frame_v) for frame_v in basis]))

    traj_unitcell_lengths = np.array(traj_unitcell_lengths)

    traj_unitcell_angles = []
    for vs in traj_box_vectors:

        angles = np.array([np.degrees(
                            np.arccos(np.dot(vs[i], vs[j])/
                                      (np.linalg.norm(vs[i]) * np.linalg.norm(vs[j]))))
                           for i, j in [(0,1), (1,2), (2,0)]])

        traj_unitcell_angles.append(angles)

    traj_unitcell_angles = np.array(traj_unitcell_angles)

    return traj_unitcell_lengths, traj_unitcell_angles

def box_vectors_to_lengths_angles(box_vectors):
    """Convert box vectors for a single 'frame' to lengths and angles.

    Parameters
    ----------
    box_vectors : arraylike of float of shape (n_dims, n_dims)
        Box vector matrices (x, y, z) for a single frame.

    Returns
    -------
    box_lengths : arraylike of float of shape (n_dims,)
        The lengths of the box

    box_angles : arraylike of float of shape (n_dims,)
        The angles of the box vectors to normal in degrees.

    """

    # calculate the lengths of the vectors through taking the norm of
    # them
    unitcell_lengths = []
    for basis in box_vectors:
        unitcell_lengths.append(np.linalg.norm(basis))
    unitcell_lengths = np.array(unitcell_lengths)

    # calculate the angles for the vectors
    unitcell_angles = np.array([np.degrees(
                        np.arccos(np.dot(box_vectors[i], box_vectors[j])/
                                  (np.linalg.norm(box_vectors[i]) * np.linalg.norm(box_vectors[j]))))
                       for i, j in [(0,1), (1,2), (2,0)]])

    return unitcell_lengths, unitcell_angles



# License applicable to the function 'lengths_and_angles_to_box_vectors'
##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

def lengths_and_angles_to_box_vectors(a_length, b_length, c_length, alpha, beta, gamma):
    """Convert from the lengths/angles of the unit cell to the box
    vectors (Bravais vectors). The angles should be in degrees.

    Parameters
    ----------
    a_length : scalar or np.ndarray
        length of Bravais unit vector **a**
    b_length : scalar or np.ndarray
        length of Bravais unit vector **b**
    c_length : scalar or np.ndarray
        length of Bravais unit vector **c**
    alpha : scalar or np.ndarray
        angle between vectors **b** and **c**, in degrees.
    beta : scalar or np.ndarray
        angle between vectors **c** and **a**, in degrees.
    gamma : scalar or np.ndarray
        angle between vectors **a** and **b**, in degrees.

    Returns
    -------
    a : np.ndarray
        If the inputs are scalar, the vectors will one dimesninoal (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)
    b : np.ndarray
        If the inputs are scalar, the vectors will one dimesninoal (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)
    c : np.ndarray
        If the inputs are scalar, the vectors will one dimesninoal (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)

    Examples
    --------
    Notes
    -----
    This code is adapted from gyroid, which is licensed under the BSD
    http://pythonhosted.org/gyroid/_modules/gyroid/unitcell.html
    >>> import numpy as np
    >>> result = lengths_and_angles_to_box_vectors(1, 1, 1, 90.0, 90.0, 90.0)
    """
    if np.all(alpha < 2*np.pi) and np.all(beta < 2*np.pi) and np.all(gamma < 2*np.pi):
        warnings.warn('All your angles were less than 2*pi. Did you accidentally give me radians?')

    alpha = alpha * np.pi / 180
    beta = beta * np.pi / 180
    gamma = gamma * np.pi / 180

    a = np.array([a_length, np.zeros_like(a_length), np.zeros_like(a_length)])
    b = np.array([b_length*np.cos(gamma), b_length*np.sin(gamma), np.zeros_like(b_length)])
    cx = c_length*np.cos(beta)
    cy = c_length*(np.cos(alpha) - np.cos(beta)*np.cos(gamma)) / np.sin(gamma)
    cz = np.sqrt(c_length*c_length - cx*cx - cy*cy)
    c = np.array([cx,cy,cz])

    if not a.shape == b.shape == c.shape:
        raise TypeError('Shape is messed up.')

    # Make sure that all vector components that are _almost_ 0 are set exactly
    # to 0
    tol = 1e-6
    a[np.logical_and(a>-tol, a<tol)] = 0.0
    b[np.logical_and(b>-tol, b<tol)] = 0.0
    c[np.logical_and(c>-tol, c<tol)] = 0.0

    return a.T, b.T, c.T


def concat_traj_fields(trajs_fields):

    # get the fields
    fields = list(trajs_fields[0].keys())

    cum_traj_fields = {}
    for field in fields:
        cum_traj_fields[field] = np.concatenate(
            [traj_fields[field] for traj_fields in trajs_fields])

    return cum_traj_fields



"""Abstract base class and null boundary condition class for
conveniently making compliant boundary condition classes for use in
wepy.

"""

import sys
import logging
from copy import deepcopy
from collections import defaultdict
import random

import numpy as np

from wepy.walker import Walker

class BoundaryConditions(object):
    """Abstract base class for conveniently making compliant boundary condition classes.

    Includes empty record group definitions and useful getters for those.

    """


    # records of boundary condition changes (sporadic)
    BC_FIELDS = ()
    """String names of fields produced in this record group.

    Boundary condition (BC) records are typically used to report on
    changes to the state of the BC object.

    Notes
    -----

    These fields are not critical to the proper functioning of the
    rest of the wepy framework and can be modified freely.

    However, reporters specific to this boundary condition probably
    will make use of these records.

    """

    BC_SHAPES = ()
    """Numpy-style shapes of all fields produced in records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A tuple of ints that specify the shape of the field element
       array.

    B. Ellipsis, indicating that the field is variable length and
       limited to being a rank one array (e.g. (3,) or (1,)).

    C. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    Note that the shapes must be tuple and not simple integers for rank-1
    arrays.

    Option B will result in the special h5py datatype 'vlen' and
    should not be used for large datasets for efficiency reasons.

    """

    BC_DTYPES = ()
    """Specifies the numpy dtypes to be used for records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A `numpy.dtype` object.

    D. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    """

    BC_RECORD_FIELDS = ()
    """Optional, names of fields to be selected for truncated
    representation of the record group.

    These entries should be strings that are previously contained in
    the 'FIELDS' class constant.

    While strictly no constraints on to which fields can be added here
    you should only choose those fields whose features could fit into
    a plaintext csv or similar format.

    """

    # warping (sporadic)
    WARPING_FIELDS = ('walker_idx', 'target_idx', 'weight')
    """String names of fields produced in this record group.

    Warping records are typically used to report whenever a walker
    satisfied the boundary conditions and was warped and had its
    state changed.

    Warnings
    --------

    Be careful when modifying these fields as they may be integrated
    with other wepy framework features. Namely recognition of
    discontinuous warping events for making contiguous trajectories
    from cloning and merging lineages.

    The behavior of whether or not a warping event is discontinuous is
    given by a `BoundaryCondition` class's `warping_discontinuity`
    which likely depends on the existence of particular fields.

    """

    WARPING_SHAPES = ((1,), (1,), (1,))
    """Numpy-style shapes of all fields produced in records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A tuple of ints that specify the shape of the field element
       array.

    B. Ellipsis, indicating that the field is variable length and
       limited to being a rank one array (e.g. (3,) or (1,)).

    C. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    Note that the shapes must be tuple and not simple integers for rank-1
    arrays.

    Option B will result in the special h5py datatype 'vlen' and
    should not be used for large datasets for efficiency reasons.

    """

    WARPING_DTYPES = (np.int, np.int, np.float)
    """Specifies the numpy dtypes to be used for records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A `numpy.dtype` object.

    D. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    """

    WARPING_RECORD_FIELDS = ('walker_idx', 'target_idx', 'weight')
    """Optional, names of fields to be selected for truncated
    representation of the record group.

    These entries should be strings that are previously contained in
    the 'FIELDS' class constant.

    While strictly no constraints on to which fields can be added here
    you should only choose those fields whose features could fit into
    a plaintext csv or similar format.

    """

    # progress towards the boundary conditions (continual)
    PROGRESS_FIELDS = ()
    """String names of fields produced in this record group.

    Progress records are typically used to report on measures of
    walkers at each cycle.

    Notes
    -----

    These fields are not critical to the proper functioning of the
    rest of the wepy framework and can be modified freely.

    However, reporters specific to this boundary condition probably
    will make use of these records.

    """

    PROGRESS_SHAPES = ()
    """Numpy-style shapes of all fields produced in records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A tuple of ints that specify the shape of the field element
       array.

    B. Ellipsis, indicating that the field is variable length and
       limited to being a rank one array (e.g. (3,) or (1,)).

    C. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    Note that the shapes must be tuple and not simple integers for rank-1
    arrays.

    Option B will result in the special h5py datatype 'vlen' and
    should not be used for large datasets for efficiency reasons.

    """

    PROGRESS_DTYPES = ()
    """Specifies the numpy dtypes to be used for records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A `numpy.dtype` object.

    D. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    """

    PROGRESS_RECORD_FIELDS = ()
    """Optional, names of fields to be selected for truncated
    representation of the record group.

    These entries should be strings that are previously contained in
    the 'FIELDS' class constant.

    While strictly no constraints on to which fields can be added here
    you should only choose those fields whose features could fit into
    a plaintext csv or similar format.

    """

    def __init__(self, **kwargs):
        """Null constructor accepts and ignores any key word arguments. """

        pass

    def bc_field_names(self):
        """Access the class level FIELDS constant for this record group."""
        return self.BC_FIELDS

    def bc_field_shapes(self):
        """Access the class level SHAPES constant for this record group."""
        return self.BC_SHAPES

    def bc_field_dtypes(self):
        """Access the class level DTYPES constant for this record group."""
        return self.BC_DTYPES

    def bc_fields(self):
        """Returns a list of zipped field specs.

        Returns
        -------

        record_specs : list of tuple
            A list of the specs for each field, a spec is a tuple of
            type (field_name, shape_spec, dtype_spec)
        """
        return list(zip(self.bc_field_names(),
                   self.bc_field_shapes(),
                   self.bc_field_dtypes()))

    def bc_record_field_names(self):
        """Access the class level RECORD_FIELDS constant for this record group."""
        return self.BC_RECORD_FIELDS

    def warping_field_names(self):
        """Access the class level FIELDS constant for this record group."""
        return self.WARPING_FIELDS

    def warping_field_shapes(self):
        """Access the class level SHAPES constant for this record group."""
        return self.WARPING_SHAPES

    def warping_field_dtypes(self):
        """Access the class level DTYPES constant for this record group."""
        return self.WARPING_DTYPES

    def warping_fields(self):
        """Returns a list of zipped field specs.

        Returns
        -------

        record_specs : list of tuple
            A list of the specs for each field, a spec is a tuple of
            type (field_name, shape_spec, dtype_spec)
        """
        return list(zip(self.warping_field_names(),
                   self.warping_field_shapes(),
                   self.warping_field_dtypes()))

    def warping_record_field_names(self):
        """Access the class level RECORD_FIELDS constant for this record group."""
        return self.WARPING_RECORD_FIELDS

    def progress_field_names(self):
        """Access the class level FIELDS constant for this record group."""
        return self.PROGRESS_FIELDS

    def progress_field_shapes(self):
        """Access the class level SHAPES constant for this record group."""
        return self.PROGRESS_SHAPES

    def progress_field_dtypes(self):
        """Access the class level DTYPES constant for this record group."""
        return self.PROGRESS_DTYPES

    def progress_fields(self):
        """Returns a list of zipped field specs.

        Returns
        -------

        record_specs : list of tuple
            A list of the specs for each field, a spec is a tuple of
            type (field_name, shape_spec, dtype_spec)
        """
        return list(zip(self.progress_field_names(),
                   self.progress_field_shapes(),
                   self.progress_field_dtypes()))

    def progress_record_field_names(self):
        """Access the class level RECORD_FIELDS constant for this record group."""
        return self.PROGRESS_RECORD_FIELDS

    def warp_walkers(self, walkers, cycle):
        """Apply boundary condition logic to walkers.

        If walkers satisfy the boundary conditions then they will be
        'warped' and have a corresponding state change take
        place. Each event recorded is returned as a single
        dictionary-style record in 'warp_data' list. These records
        correspond to the 'WARPING' record group.

        Additional data calculated on walkers may be returned in the
        single 'progress_data' dictionary-style record, which
        corresponds to the 'PROGRESS' record group.

        Any changes to the internal state of the boundary condition
        object (e.g. modification of parameters) should be recorded in
        at least one dictionary-style record in the 'bc_data'
        list. This corresponds to the 'BC' record group.

        Parameters
        ----------
        walkers : list of walkers
            A list of objects implementing the Walker interface

        cycle : int
            The index of the cycle this is for. Used to generate proper records.

        Returns
        -------
        new_walkers : list of walkers
            A list of objects implementing the Walker interface, that have had
            boundary condition logic applied.

        warp_data : list of dict of str : value
            A list of dictionary style records for each warping event that occured.

        bc_data : list of dict of str : value
            A list of dictionary style records for each boundary condition state change
            event record that occured.

        progress_data : dict of str : list of value
           Dictionary style progress records. The values should be lists
           corresponding to each walker.

        """

        raise NotImplementedError

    @classmethod
    def warping_discontinuity(cls, warping_record):
        """Given a warping record returns either True for a discontiuity
        occured or False if a discontinuity did not occur.

        Parameters
        ----------
        warping_record : tuple
            A tuple record of type 'WARPING'

        Returns
        -------
        is_discontinuous : bool
            True if discontinuous warping record False if continuous.

        """

        raise NotImplementedError



class NoBC(BoundaryConditions):
    """Boundary conditions class that does nothing.

    You may use this class as a stub in order to have an boundary
    condition class. However, this is not necessary since boundary
    conditions are optional in the sim_manager anyhow.
    """

    def warp_walkers(self, walkers, cycle):
        """Apply boundary condition logic to walkers, of which there is none.

        Simply returns all walkers provided with empty records data
        since there is nothing to do.

        Parameters
        ----------
        walkers : list of walkers
            A list of objects implementing the Walker interface

        cycle : int
            The index of the cycle this is for. Used to generate proper records.

        Returns
        -------
        new_walkers : list of walkers
            A list of objects implementing the Walker interface, that have had
            boundary condition logic applied.

        warp_data : list of dict of str : value
            A list of dictionary style records for each warping event that occured.

        bc_data : list of dict of str : value
            A list of dictionary style records for each boundary condition state change
            event record that occured.

        progress_data : dict of str : list of value
           Dictionary style progress records. The values should be lists
           corresponding to each walker.

        """

        warp_data = []
        bc_data = []
        progress_data = {}

        return walkers, warp_data, bc_data, progress_data

    @classmethod
    def warping_discontinuity(cls, warping_record):
        # documented in superclass

        # always return false
        return False


class RandomBC(BoundaryConditions):
    """Boundary conditions that randomly warps both continuously and
    discontinuously.

    Can be used with any system as it won't actually mutate states.

    """


    # records of boundary condition changes (sporadic)
    BC_FIELDS = ('ping',)
    BC_SHAPES = ((1,),)
    BC_DTYPES = (np.int,)

    BC_RECORD_FIELDS = ('ping',)

    # warping fields are directly inherited

    # progress towards the boundary conditions (continual)
    PROGRESS_FIELDS = ('weight',)
    PROGRESS_SHAPES = (Ellipsis,)
    PROGRESS_DTYPES = (np.float,)

    PROGRESS_RECORD_FIELDS = ('weight',)

    DISCONTINUITY_TARGET_IDXS = (0,)

    def warp_walkers(self, walkers, cycle):

        ## warping walkers

        # just return the same walkers
        new_walkers = deepcopy(walkers)

        ## warping data

        warp_data = []
        # generate warping data: 50% of the time generate a warping
        # event, 25% is discontinuous (target 0), and 25% is
        # continuous (target 1)
        for walker_idx, walker in enumerate(walkers):

            # warping event?
            if random.random() >= 0.5:

                # discontinuous?
                if random.random() >= 0.5:
                    warp_record = {
                        'walker_idx' : np.array([walker_idx]),
                        'target_idx' : np.array([0]),
                        'weight' : np.array([walker.weight]),
                    }

                    warp_data.append(warp_record)

                # continuous
                else:
                    warp_record = {
                        'walker_idx' : np.array([walker_idx]),
                        'target_idx' : np.array([1]),
                        'weight' : np.array([walker.weight]),
                    }

                    warp_data.append(warp_record)

        ## BC data
        bc_data = []
        # choose whether to generate a bc record
        if random.random() >= 0.5:
            bc_data.append({'ping' : np.array([1])})

        ## Progress data

        # just set the walker progress to be its weight so there is a
        # number there
        progress_data = {'weight' :
                         [walker.weight for walker in walkers]
        }


        return new_walkers, warp_data, bc_data, progress_data




"""This is a exmaple of conventional random walke simultion. There is
no resampling after the random walk dynamics.

"""
import sys
import os
import os.path as osp
from pathlib import Path

import numpy as np

from wepy.resampling.resamplers.resampler import NoResampler
from wepy.resampling.distances.randomwalk import RandomWalkDistance
from wepy.runners.randomwalk import RandomWalkRunner, UNIT_NAMES
from wepy.walker import Walker, WalkerState

from wepy_tools.sim_makers.toys.randomwalk import RandomwalkProfiler



SAVE_FIELDS = ('positions')
# Name of field's unit in the HDF5
UNITS = UNIT_NAMES

outputs_dir = Path('_output')

if not osp.exists(outputs_dir):
    os.makedirs(outputs_dir)

# sets the input paths
hdf5_filename = 'rw_results.wepy.h5'
reporter_filename = 'randomwalk_conventional.org'

hdf5_path= outputs_dir / hdf5_filename
reporter_path = outputs_dir / reporter_filename



if __name__=="__main__":
    if sys.argv[1] == "--help" or sys.argv[1] == '-h':
        print("arguments: n_cycles, n_walkers, dimension")
    else:
        n_runs = int(sys.argv[1])
        n_cycles = int(sys.argv[2])
        n_walkers = int(sys.argv[3])
        dimension = int(sys.argv[4])

    # set up  the distance function
    distance = RandomWalkDistance();



    # set up the NOResampler
    resampler = NoResampler()

    # set up a RandomWalkProfilier
    rw_profiler = RandomwalkProfiler(resampler,
                                     dimension,
                                     hdf5_filename=str(hdf5_path),
                                     reporter_filename=str(reporter_path))

    # runs the simulations and gets the result
    rw_profiler.run(
        num_runs=n_runs,
        num_cycles=n_cycles,
        num_walkers=n_walkers,
    )



"""Analysis routines for generating free energy profiles over wepy
simulation data.

"""
import gc
import itertools as it

import numpy as np

from geomm.free_energy import free_energy

def _percentage_cumulative_partitions(ensemble_values,
                                     percentages=[0.25, 0.5, 0.75, 1.0]):

    n_cycles = ensemble_values.shape[0]
    n_trajs = ensemble_values.shape[1]

    start_idx = 0
    partitions = []
    for percentage in percentages:

        # get the percentage of the cycles based on this percentage
        end_idx = int(n_cycles // (1 / percentage))

        partitions.append((start_idx, end_idx))

    return partitions


def _tranche_cumulative_partitions(ensemble_values,
                                  time_tranche=None,
                                  num_partitions=4):
    n_cycles = ensemble_values.shape[0]
    n_trajs = ensemble_values.shape[1]

    start_idx = 0

    # if time_tranche was given determine the number of partitions
    if time_tranche is not None and time_tranche > 0:
        num_partitions = n_cycles // time_tranche

        # get the end indices for each tranche
        end_idxs = []
        for i in range(num_partitions):
            end_idx = time_tranche * (i + 1)

        # and create the array of partitions
        partitions = [(start_idx, end_idx) for end_idx in end_idxs]

        # if there is a remainder we will have another partition for
        # the remainder
        remainder = n_cycles % time_tranche
        if remainder > 0:
            num_partitions += 1
            end_idx = remainder + (tim_tranche * num_partitions)

            partitions.append((start_idx, end_idx))

    # check that one of them is defined
    elif num_partitions is None and num_partitions > 0:
        raise ValueError("Must specify either time_tranche size or partitions ratio")

    # if the partitions are given and the time_tranche is not we
    # determine the tranche size automatically
    else:
        time_tranche = n_cycles // num_partitions

        end_idxs = []
        for i in range(num_partitions):
            end_idx = time_tranche * (i + 1)

        # and create the array of partitions
        partitions = [(start_idx, end_idx) for end_idx in end_idxs]

        # if there is a remainder in this case we add it to the first
        # partition since the number of partitions is invariant
        # (unlike time_tranche option)
        remainder = n_cycles % time_tranche
        if remainder > 0:
            partitions[0][1] += remainder

    return partitions

# REFACTOR: this has an overly complex decision process encapsulated
# here, should be refactored. Was not properly done because of time
# constraints
def cumulative_partitions(ensemble_values,
                          time_tranche=None,
                          num_partitions=5,
                          percentages=None):
    """Calculate the cumulative free energies for an ensemble of
    trajectory weights.

    Parameters
    ----------

    ensemble_values : arraylikes of float of shape (n_cycles, n_trajs)
        Array of scalar values for all of the frames of an ensemble simulation.

    time_tranche : int
        Width of time-bins/windows/tranche to compute free energies
        for. This overrides the functionality of the 'partitions'
        argument. The last tranche will be truncated to the remainder
        after partitioning by tranches.

    num_partitions : int
        Will evenly partition the time coordinate given the
        dataset. The first partition will carry the
        remainder. Overriden by time_tranche.

    percentages: list of float, optional

    Yields
    ------

    cumulative_tranche : arraylike
        A slice along the cycles axis of the ensemble values starting
        at the beginning and including up to the end of the next
        tranche.

    """


    if percentages is None:
        partitions = _tranche_cumulative_partitions(ensemble_values,
                                                    time_tranche=time_tranche,
                                                    num_partitions=num_partitions)

    else:
        partitions = _percentage_cumulative_partitions(ensemble_values,
                                                       percentages=percentages)

    # yield the time-partitioned values when asked for
    for start_idx, end_idx in partitions:

        # compute the free energies
        yield ensemble_values[start_idx:end_idx]

def free_energy_profile(weights, observables, bins=30,
                  max_energy=100,
                  zero_point_energy=1e-12):
    """
    Parameters
    ----------

    weights : arraylikes of float of shape (n_trajs, n_cycles)
        The weights for all of the frames of an ensemble simulation.

    observables : arraylikes of shape (n_trajs, n_cycles)
        The scalar observables to compute the free energy over.

    bins : int or sequence of scalars or str
        The number of bins to bin along the observable axis, or the
        actual bin edges, or the spec for the binning method.

    Returns
    -------

    hist_fe : arraylike
        The free energies of the histogram bins

    bin_edges : arraylike
        The edges of the bins of the histogram.

    """

    assert weights.shape == observables.shape, "Weights and observables must correspond in shape"

    hist_weights, bin_edges = np.histogram(observables, weights=weights, bins=bins)

    hist_fe = free_energy(hist_weights,
                          max_energy=np.nan,
                          zero_point_energy=zero_point_energy)

    return hist_fe, bin_edges

def contigtrees_bin_edges(contigtrees, bins, field_key,
                          truncate_cycles=None):
    """Get the bin edges that best describes the aggregate data for
    multiple contigtrees.

    Parameters
    ----------

    contigtrees : list of ContigTree objects
        The contigtrees to draw the data from.

    bins : int or str or sequence of scalars or function
        The number of bins to make or the method to use for
        autobinning or the actual specification of bins. See
        `numpy.histogram_bin_edges` for exact meaning and methods.

    field_key : str
        The trajectory field key for the values to get bins for.

    truncate_cycles : None or int
        If None include all cycles of a run. Otherwise only consider
        cycles for spans up to (exclusive) the cycle index given.

    Returns
    -------
    bin_edges : array of dtype float
        The edges to pass into `numpy.histogram`


    See Also
    --------
    numpy.histogram_bin_edges : For the bin edge function

    """

    all_values = []

    for contigtree in contigtrees:
        with contigtree:

            contigtree_values = []
            for span_idx in contigtree.span_traces.keys():

                contig = contigtree.span_contig(span_idx)

                if truncate_cycles is not None:
                    contig_values = contig.contig_fields([field_key])\
                                    [field_key][0:truncate_cycles]

                else:
                    contig_values = contig.contig_fields([field_key])[field_key]

                contigtree_values.append(contig_values.reshape(-1))

            all_values.extend(contigtree_values)

    all_values = np.concatenate(all_values)
    gc.collect()

    if callable(bins):
        bin_edges = bins(all_values)

    else:
        bin_edges = np.histogram_bin_edges(all_values, bins=bins)

    # If the histogram_bin_edges actually used the weights we would
    # call it like this and actually have to retrieve the weights,
    # currently it doesn't so I won't implement that since it would
    # slow things down and not do anything

    # bin_edges = np.histogram_bin_edges(all_values, bins=bins,
    #                                    weights=all_weights)

    return bin_edges


class ContigTreeProfiler(object):
    """A wrapper class around a ContigTree that provides extra methods for
    generating free energy profiles for observables."""

    def __init__(self, contigtree,
                 truncate_cycles=None):
        """Create a wrapper around a contigtree for the profiler.

        Parameters
        ----------

        contigtree : ContigTree object
            The contigtree you want to generate profiles for.

        truncate_cycles : None or int
            If None include all cycles of a run. Otherwise only consider
            cycles for spans up to (exclusive) the cycle index given.

        """

        self._contigtree = contigtree

        self._truncate_cycles = truncate_cycles

        # get the ignored frames trace as a set
        if truncate_cycles is not None:
            self._ignore_trace = self._get_ignore_trace(self._contigtree, truncate_cycles)
        else:
            self._ignore_trace = set()

        # determine the global binning strategy. These are the bins
        # that best describe the entire pool of data in this contig
        # tree

    @classmethod
    def _get_ignore_trace(cls, contigtree, truncate_cycles):
        """Get a trace of the ignore frames if we are truncating on a
        particular cycle index. Is exclusive of the `truncate_cycles`
        value.

        Parameters
        ----------

        contigtree : ContigTree

        truncate_cycles : int
            The cycle index to truncate at (exclusive).

        Returns
        -------

        ignored_trace : set of (int, int)
            The frames to ignore given the truncation.

        """

        ignore_trace = set()
        for span_idx, span_trace in contigtree.span_traces.items():

            for run_idx, cycle_idx in span_trace:
                if cycle_idx >= truncate_cycles:
                    ignore_trace.add((run_idx, cycle_idx,))

        return ignore_trace

    @property
    def contigtree(self):
        """The underlying contigtree this wraps."""
        return self._contigtree

    def fe_profile_trace(self, trace, field_key,
                         bins=None,
                         ignore_truncate=False):
        """Calculate the free energy histogram over a trajectory field.

        Parameters
        ----------

        trace : list of tuple of ints (run_idx, traj_idx, cycle_idx)

        field_key : str
            The key for the trajectory field to calculate the profiles
            for. Must be a rank 0 (or equivalent (1,) rank) field.

        bins : int or sequence of scalars or str or None
            The number of bins to bin along the observable axis, or
            the actual bin edges, or the spec for the binning method,
            or None. If None will generate them with the 'auto'
            setting for this span.

        ignore_truncate : bool
            Ignore the truncate_cycles value.

        Returns
        -------

        fe_profile : arraylike of dtype float
            An array of free energies for each bin.

        """


        if not ignore_truncate:
            ignored_trace = []
            for key in trace:
                if key not in self._ignore_trace:
                    ignored_trace.append(key)

            trace = ignored_trace

        fields = self.contigtree.wepy_h5.get_trace_fields(trace, [field_key, 'weights'])

        weights = fields['weights'].flatten()
        values = fields[field_key]

        # the feature vector must be either shape 0 or (1,) i.e. rank
        # 0 equivalent, so we get one feature and check its shape
        test_feature = values[0]

        # if it doesn't satisfy this requirement we need to raise an error
        if not (test_feature.shape == () or test_feature.shape == (1,)):
            raise TypeError("The field key specified a field that is non-scalar."
                            "The shape of the feature vector must be () or (1,).")

        # determine the bin_edges

        # if the bins were not specified generate them from all the
        # values for this span
        if bins is None:
            bin_edges = np.histogram_bin_edges(values, bins='auto',
                                               weights=weights)

        # if the number or binning strategy was given generate them
        # according to that
        elif type(bins) in (str, int):
            bin_edges = np.histogram_bin_edges(values, bins=bins,
                                               weights=weights)

        # if it wasn't those things we assume it is already a correct
        # bin_edges
        else:
            bin_edges = bins


        fe_profile, _ = free_energy_profile(weights, values,
                                            bins=bin_edges)

        return fe_profile


    def fe_profile_all(self, field_key, bins=None,
                       ignore_truncate=False):
        """Calculate the free energy histogram over a trajectory field.

        Parameters
        ----------

        field_key : str
            The key for the trajectory field to calculate the profiles
            for. Must be a rank 0 (or equivalent (1,) rank) field.

        bins : int or sequence of scalars or str or None
            The number of bins to bin along the observable axis, or
            the actual bin edges, or the spec for the binning method,
            or None. If None will generate them with the 'auto'
            setting for this span.

        ignore_truncate : bool
            Ignore the truncate_cycles value.


        Returns
        -------

        fe_profile : arraylike of dtype float
            An array of free energies for each bin.

        """

        all_trace = list(it.chain(*[span_trace for span_trace in self.contigtree.span_traces.values()]))

        # this truncates based on the frame specs so doesn't need to
        # be ordered or anything
        fe_profile = self.fe_profile_trace(all_trace, field_key,
                                           bins=bins,
                                           ignore_truncate=ignore_truncate)

        return fe_profile


    def fe_profile(self, span, field_key,
                   bins=None,
                   ignore_truncate=False):
        """Calculate the free energy histogram over a trajectory field.

        Deprecation warning: This will likely be deprecated and
        renamed to fe_profile_span because it is misleading that a
        profile is span based.

        Parameters
        ----------

        span : int
            The index of the span to calculate profiles for.

        field_key : str
            The key for the trajectory field to calculate the profiles
            for. Must be a rank 0 (or equivalent (1,) rank) field.

        bins : int or sequence of scalars or str or None
            The number of bins to bin along the observable axis, or
            the actual bin edges, or the spec for the binning method,
            or None. If None will generate them with the 'auto'
            setting for this span.

        ignore_truncate : bool
            Ignore the truncate_cycles value.


        Returns
        -------

        fe_profile : arraylike of dtype float
            An array of free energies for each bin.

        """

        # make the contig for this span
        contig = self.contigtree.span_contig(span)

        if not ignore_truncate:
            # get the weights
            weights = contig.contig_fields(['weights'])['weights'][0:self._truncate_cycles]

            # then get the values for the field key
            values = contig.contig_fields([field_key])[field_key][0:self._truncate_cycles]

        else:
            # get the weights
            weights = contig.contig_fields(['weights'])['weights'][0:self._truncate_cycles]

            # then get the values for the field key
            values = contig.contig_fields([field_key])[field_key][0:self._truncate_cycles]

        # reshape to match
        weights = weights.reshape((weights.shape[0], weights.shape[1]))

        # the feature vector must be either shape 0 or (1,) i.e. rank
        # 0 equivalent

        # so we get one feature and check its shape
        test_feature = values[0,0]

        # if it doesn't satisfy this requirement we need to raise an error
        if not (test_feature.shape == () or test_feature.shape == (1,)):
            raise TypeError("The field key specified a field that is non-scalar."
                            "The shape of the feature vector must be () or (1,).")

        # determine the bin_edges

        # if the bins were not specified generate them from all the
        # values for this span
        if bins is None:
            bin_edges = np.histogram_bin_edges(values, bins='auto',
                                               weights=weights)

        # if the number or binning strategy was given generate them
        # according to that
        elif type(bins) in (str, int):
            bin_edges = np.histogram_bin_edges(values, bins=bins,
                                               weights=weights)

        # if it wasn't those things we assume it is already a correct
        # bin_edges
        else:
            bin_edges = bins


        fe_profile, _ = free_energy_profile(weights, values,
                                            bins=bin_edges)

        return fe_profile


    def fe_cumulative_profiles(self, span, field_key,
                               bins=None,
                               time_tranche=None,
                               num_partitions=5,
                               ignore_truncate=False):
        """Calculate the cumulative free energy histograms for a trajectory
        field.

        Parameters
        ----------

        span : int
            The index of the span to calculate profiles for.

        field_key : str
            The key for the trajectory field to calculate the profiles
            for. Must be a rank 0 (or equivalent (1,) rank) field.

        bins : int or sequence of scalars or str or None
            The number of bins to bin along the observable axis, or
            the actual bin edges, or the spec for the binning method,
            or None. If None will generate them with the 'auto'
            setting for this span.

        time_tranche : int
            Width of time-bins/windows/tranche to compute free
            energies for. This overrides the functionality of the
            'num_partitions' argument. The last tranche will be truncated
            to the remainder after partitioning by tranches. Overrides
            partitions.

        num_partitions : int
            Will evenly partition the time coordinate given the
            dataset. The first partition will carry the remainder.
            Overriden by time_tranche.

        ignore_truncate : bool
            Ignore the truncate_cycles value.

        Returns
        -------

        cumulative_fe_profiles : list of arraylike of dtype float
            A list of each cumulative free energy profile. Each profile is
            an array of free energies for each bin.

        profile_num_cycles : list of int
            The number of cycles in each cumulative fe profile.

        """

        # make the contig for this span
        contig = self.contigtree.span_contig(span)

        if not ignore_truncate:
            # get the weights
            weights = contig.contig_fields(['weights'])['weights'][0:self._truncate_cycles]
            # then get the values for the field key
            values = contig.contig_fields([field_key])[field_key][0:self._truncate_cycles]

        else:
            # get the weights
            weights = contig.contig_fields(['weights'])['weights']
            # then get the values for the field key
            values = contig.contig_fields([field_key])[field_key]

        # reshape to match
        weights = weights.reshape((weights.shape[0], weights.shape[1]))

        # determine the bin_edges

        # if the bins were not specified generate them from all the
        # values for this span
        if bins is None:
            bin_edges = np.histogram_bin_edges(values, bins='auto',
                                               weights=weights)

        # if the number or binning strategy was given generate them
        # according to that
        elif type(bins) in (str, int):
            bin_edges = np.histogram_bin_edges(values, bins=bins,
                                               weights=weights)

        # if it wasn't those things we assume it is already a correct
        # bin_edges
        else:
            bin_edges = bins


        # make the cumulative generators for each
        weights_partition_gen = cumulative_partitions(weights, time_tranche=time_tranche,
                                                      num_partitions=num_partitions)
        values_partition_gen = cumulative_partitions(values, time_tranche=time_tranche,
                                                     num_partitions=num_partitions)

        # then generate each profile for each cumulative tranche

        fe_profiles = []
        profile_num_cycles = []
        for weights_part in weights_partition_gen:

            values_part = next(values_partition_gen)

            profile_num_cycles.append(len(values_part))

            fe_profile, _ = free_energy_profile(weights_part, values_part,
                                                bins=bin_edges)

            fe_profiles.append(fe_profile)

        return fe_profiles, profile_num_cycles


    def fe_all_cumulative_profiles(self, field_key,
                                   bins=None,
                                   time_tranche=None,
                                   num_partitions=5,
                                   percentages=None,
                                   ignore_truncate=False):
        """Calculate the cumulative free energy histograms for a trajectory
        field for all spans in a contigtree.

        Parameters
        ----------

        field_key : str
            The key for the trajectory field to calculate the profiles
            for. Must be a rank 0 (or equivalent (1,) rank) field.

        bins : int or sequence of scalars or str or None
            The number of bins to bin along the observable axis, or
            the actual bin edges, or the spec for the binning method,
            or None. If None will generate them with the 'auto'
            setting for this span.

        time_tranche : int
            Width of time-bins/windows/tranche to compute free
            energies for. This overrides the functionality of the
            'num_partitions' argument. The last tranche will be truncated
            to the remainder after partitioning by tranches. Overrides
            partitions.

        num_partitions : int
            Will evenly partition the time coordinate given the
            dataset. The first partition will carry the remainder.
            Overriden by time_tranche.

        percentages : list of float

        ignore_truncate : bool
            Ignore the truncate_cycles value.

        Returns
        -------

        cumulative_fe_profiles : list of arraylike of dtype float
            A list of each cumulative free energy profile. Each profile is
            an array of free energies for each bin.

        profile_num_cycles : list of int
            The number of cycles in each cumulative fe profile.

        """


        # There are two big use cases here.
        #
        # 1. We are passing in the bin edges since we want to plot
        # this profile with other data on the same scale
        #
        # 2. We are just wanting to show this one on its own.
        #
        # In the former we don't do anything and fail if it doesn't
        # work. In the latter we want to infer bins like normal. The
        # first thing we do is distinguish between which case we have.

        # if the bins were not specified generate them from all the
        # values for this span
        if bins is None or type(bins) in (str, int):

            # if we are doing the former we need to get the bins for
            # all of the data first. Once we have that we worry about
            # doing the tranches etc. So get all the data, get the
            # bins, and then delete it and start fresh with the
            # tranches

            # we use a trace orientation here for simplicity in truncation

            # trace of all of the frames in the contigtree
            all_trace = list(it.chain(*[span_trace for span_trace in
                                        self.contigtree.span_traces.values()]))

            # filter it for the ignored fields if applicable
            if not ignore_truncate:
                ignored_trace = []
                for key in all_trace:
                    if key not in self._ignore_trace:
                        ignored_trace.append(key)

                trace = ignored_trace

            else:
                trace = all_trace

            # get the fields for the weights and the field
            fields = self.contigtree.wepy_h5.get_trace_fields(trace, [field_key, 'weights'])


            all_weights = fields['weights']
            all_values = fields[field_key]

            all_weights = all_weights.flatten()
            gc.collect()

            # the feature vector must be either shape 0 or (1,) i.e. rank
            # 0 equivalent

            # so we get one feature and check its shape
            test_feature = all_values[0]

            # if it doesn't satisfy this requirement we need to raise an error
            if not (test_feature.shape == () or test_feature.shape == (1,)):
                raise TypeError("The field key specified a field that is non-scalar."
                                "The shape of the feature vector must be () or (1,).")

            # then get the bin edges according to the specific option
            if bins is None:
                bin_edges = np.histogram_bin_edges(all_values, bins='auto',
                                                   weights=all_weights)

            # if the number or binning strategy was given generate them
            # according to that
            elif type(bins) in (str, int):
                bin_edges = np.histogram_bin_edges(all_values, bins=bins,
                                                   weights=all_weights)

            # clean up these values since we will be loading in new
            # values that are structured in a different way.
            del all_values
            del all_weights
            gc.collect()

        # if it wasn't those things we assume it is already a correct
        # bin_edges
        else:
            bin_edges = bins

        # now we need to decide on the length of the interval that we
        # will be using so that we can partition it and then use that
        # to grab the right subset of data

        spans_weights_part_gens = []
        spans_values_part_gens = []
        for span_idx in self.contigtree.span_traces.keys():

            contig = self.contigtree.span_contig(span_idx)

            if not ignore_truncate:
                # get the weights
                contig_weights = contig.contig_fields(['weights'])['weights'][0:self._truncate_cycles]
                # then get the values for the field key
                contig_values = contig.contig_fields([field_key])[field_key][0:self._truncate_cycles]

            else:
                # get the weights
                contig_weights = contig.contig_fields(['weights'])['weights']
                # then get the values for the field key
                contig_values = contig.contig_fields([field_key])[field_key]

            # reshape to match
            contig_weights = contig_weights.reshape((contig_weights.shape[0],
                                                     contig_weights.shape[1]))

            # make the cumulative generators for each
            contig_weights_partition_gen = cumulative_partitions(contig_weights,
                                                          time_tranche=time_tranche,
                                                                 num_partitions=num_partitions,
                                                                 percentages=percentages)
            contig_values_partition_gen = cumulative_partitions(contig_values,
                                                         time_tranche=time_tranche,
                                                                num_partitions=num_partitions,
                                                                percentages=percentages)

            # we want to include all spans at once so save these for the next part
            spans_weights_part_gens.append(contig_weights_partition_gen)
            spans_values_part_gens.append(contig_values_partition_gen)


        part_profiles = []
        # use the first one for the main loop iteration
        for first_weights_part in spans_weights_part_gens[0]:

            gc.collect()

            # then add the other ones
            weights_parts = [first_weights_part.reshape(-1)]
            for other_weights_part_gen in spans_weights_part_gens[1:]:

                weights_part = next(other_weights_part_gen)
                weights_parts.append(weights_part.reshape(-1))

            values_parts = []
            for values_part_gen in spans_values_part_gens:

                values_part = next(values_part_gen)
                values_parts.append(values_part.reshape(-1))

            # concatenate them to feed them to the main function
            agg_weights_part = np.concatenate(weights_parts)
            agg_values_part = np.concatenate(values_parts)

            # then compute the FE curve from the aggregate of spans
            part_profile, _ = free_energy_profile(agg_weights_part,
                                                  agg_values_part,
                                                  bins=bin_edges)

            part_profiles.append(part_profile)

        return part_profiles

    def bin_edges(self, bins, field_key):
        """Compute the bin edges that best describes all of the contigtree
        data for the number of bins.

        Parameters
        ----------

        bins : int or str
            The number of bins to make or the method to use for
            autobinning.

        field_key : str
            The trajectory field key for the values to get bins for.

        Returns
        -------
        bin_edges : array of dtype float
            The edges to pass into `numpy.histogram`


        See Also
        --------
        numpy.histogram_bin_edges : For the bin edge function

        """

        all_values = np.concatenate([fields[field_key]
                                     for fields
                                     in self.contigtree.wepy_h5.iter_trajs_fields([field_key])])

        bin_edges = np.histogram_bin_edges(all_values, bins=bins)

        # If the histogram_bin_edges actually used the weights we would
        # call it like this and actually have to retrieve the weights,
        # currently it doesn't so I won't implement that since it would
        # slow things down and not do anything

        # bin_edges = np.histogram_bin_edges(all_values, bins=bins,
        #                                    weights=all_weights)


        return bin_edges

